try:
    from enum import StrEnum  # type: ignore
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):
        """Compatible with python below 3.11"""


from pathlib import Path
from tempfile import gettempdir
from typing import Any, Literal, Optional

from rich.console import JustifyMethod

BOOL_STR = Literal["true", "false", "yes", "no", "y", "n", "1", "0", "on", "off"]


class JustifyEnum(StrEnum):  # type: ignore
    DEFAULT = "default"
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"
    FULL = "full"


CMD_CLEAR = "/clear"
CMD_EXIT = "/exit"
CMD_HISTORY = "/his"
CMD_MODE = "/mode"
CMD_SAVE_CHAT = "/save"
CMD_LOAD_CHAT = "/load"
CMD_LIST_CHATS = "/list"
CMD_DELETE_CHAT = "/del"
CMD_CONTEXT = ("/context", "/ctx")
CMD_ADD = "/add"
CMD_HELP = ("/help", "?")

EXEC_MODE = "exec"
CHAT_MODE = "chat"
TEMP_MODE = "temp"
CODE_MODE = "code"

HISTORY_FILE = Path("~/.yaicli_history").expanduser()
CONFIG_PATH = Path("~/.config/yaicli/config.ini").expanduser()
ROLES_DIR = CONFIG_PATH.parent / "roles"
FUNCTIONS_DIR = CONFIG_PATH.parent / "functions"
MCP_JSON_PATH = CONFIG_PATH.parent / "mcp.json"

# Default configuration values
DEFAULT_CODE_THEME = "monokai"
DEFAULT_PROVIDER = "openai"
DEFAULT_BASE_URL = "https://api.openai.com/v1"
DEFAULT_MODEL = "gpt-4o"
DEFAULT_SHELL_NAME = "auto"
DEFAULT_ROLE = "DEFAULT"
DEFAULT_OS_NAME = "auto"
DEFAULT_STREAM: BOOL_STR = "true"
DEFAULT_TEMPERATURE: float = 0.3
DEFAULT_FREQUENCY_PENALTY: float = 0.0
DEFAULT_TOP_P: float = 1.0
DEFAULT_MAX_TOKENS: int = 1024
DEFAULT_MAX_HISTORY: int = 500
DEFAULT_AUTO_SUGGEST: BOOL_STR = "true"
DEFAULT_SHOW_REASONING: BOOL_STR = "true"
DEFAULT_TIMEOUT: int = 60
DEFAULT_EXTRA_HEADERS: str = "{}"
DEFAULT_EXTRA_BODY: str = "{}"
DEFAULT_INTERACTIVE_ROUND: int = 25
DEFAULT_CHAT_HISTORY_DIR: Path = Path(gettempdir()) / "yaicli/chats"
DEFAULT_MAX_SAVED_CHATS = 20
DEFAULT_JUSTIFY: JustifyMethod = "default"
DEFAULT_ROLE_MODIFY_WARNING: BOOL_STR = "true"
DEFAULT_ENABLE_FUNCTIONS: BOOL_STR = "true"
DEFAULT_SHOW_FUNCTION_OUTPUT: BOOL_STR = "true"
# low/high/medium for openai, default/null for groq
DEFAULT_REASONING_EFFORT: Optional[Literal["low", "high", "medium", "default", "null"]] = None
DEFAULT_ENABLE_MCP: BOOL_STR = "false"
DEFAULT_SHOW_MCP_OUTPUT: BOOL_STR = "false"
DEFAULT_MAX_TOOL_CALL_DEPTH: int = 8
DEFAULT_EXCLUDE_PARAMS: str = ""  # Empty by default


SHELL_PROMPT = """You are YAICLI, a shell command generator.
The context conversation may contain other types of messages, 
but you should only respond with a single valid {_shell} shell command for {_os}.
Do not include any explanations, comments, or formatting — only the command as plain text, avoiding Markdown formatting.
"""

DEFAULT_PROMPT = """
You are YAICLI, a system management and programing assistant, 
You are managing {_os} operating system with {_shell} shell. 
Your responses should be concise and use Markdown format (but dont't use ```markdown), 
unless the user explicitly requests more details.
"""

CODER_PROMPT = (
    "You are YAICLI, a code assistant. "
    "You are helping with programming tasks. "
    "Your responses must ONLY contain code, with NO explanation, NO markdown formatting, and NO preamble. "
    "If user does not specify the language, provide Python code. "
    "Do not wrap code in markdown code blocks (```) or language indicators."
)


class DefaultRoleNames(StrEnum):  # type: ignore
    SHELL = "Shell Command Generator"
    DEFAULT = "DEFAULT"
    CODER = "Code Assistant"


DEFAULT_ROLES: dict[str, dict[str, Any]] = {
    DefaultRoleNames.SHELL.value: {"name": DefaultRoleNames.SHELL.value, "prompt": SHELL_PROMPT},  # type: ignore
    DefaultRoleNames.DEFAULT.value: {"name": DefaultRoleNames.DEFAULT.value, "prompt": DEFAULT_PROMPT},  # type: ignore
    DefaultRoleNames.CODER.value: {"name": DefaultRoleNames.CODER.value, "prompt": CODER_PROMPT},  # type: ignore
}

# DEFAULT_CONFIG_MAP is a dictionary of the configuration options.
# The key is the name of the configuration option.
# The value is a dictionary with the following keys:
# - value: the default value of the configuration option
# - env_key: the environment variable key of the configuration option
# - type: the type of the configuration option, dict -> json, bool -> boolstr
DEFAULT_CONFIG_MAP = {
    # Core API settings
    "BASE_URL": {"value": "", "env_key": "YAI_BASE_URL", "type": str},
    "API_KEY": {"value": "", "env_key": "YAI_API_KEY", "type": str},
    "MODEL": {"value": DEFAULT_MODEL, "env_key": "YAI_MODEL", "type": str},
    # System detection hints
    "SHELL_NAME": {"value": DEFAULT_SHELL_NAME, "env_key": "YAI_SHELL_NAME", "type": str},
    "OS_NAME": {"value": DEFAULT_OS_NAME, "env_key": "YAI_OS_NAME", "type": str},
    "DEFAULT_ROLE": {"value": DEFAULT_ROLE, "env_key": "YAI_DEFAULT_ROLE", "type": str},
    # API call parameters
    "STREAM": {"value": DEFAULT_STREAM, "env_key": "YAI_STREAM", "type": bool},
    "TEMPERATURE": {"value": DEFAULT_TEMPERATURE, "env_key": "YAI_TEMPERATURE", "type": float},
    "FREQUENCY_PENALTY": {"value": DEFAULT_FREQUENCY_PENALTY, "env_key": "YAI_FREQUENCY_PENALTY", "type": float},
    "TOP_P": {"value": DEFAULT_TOP_P, "env_key": "YAI_TOP_P", "type": float},
    "MAX_TOKENS": {"value": DEFAULT_MAX_TOKENS, "env_key": "YAI_MAX_TOKENS", "type": int},
    "TIMEOUT": {"value": DEFAULT_TIMEOUT, "env_key": "YAI_TIMEOUT", "type": int},
    "EXTRA_HEADERS": {"value": DEFAULT_EXTRA_HEADERS, "env_key": "YAI_EXTRA_HEADERS", "type": dict},
    "EXTRA_BODY": {"value": DEFAULT_EXTRA_BODY, "env_key": "YAI_EXTRA_BODY", "type": dict},
    "REASONING_EFFORT": {"value": DEFAULT_REASONING_EFFORT, "env_key": "YAI_REASONING_EFFORT", "type": str},
    "INTERACTIVE_ROUND": {
        "value": DEFAULT_INTERACTIVE_ROUND,
        "env_key": "YAI_INTERACTIVE_ROUND",
        "type": int,
    },
    # UI/UX settings
    "CODE_THEME": {"value": DEFAULT_CODE_THEME, "env_key": "YAI_CODE_THEME", "type": str},
    "MAX_HISTORY": {"value": DEFAULT_MAX_HISTORY, "env_key": "YAI_MAX_HISTORY", "type": int},
    "AUTO_SUGGEST": {"value": DEFAULT_AUTO_SUGGEST, "env_key": "YAI_AUTO_SUGGEST", "type": bool},
    "SHOW_REASONING": {"value": DEFAULT_SHOW_REASONING, "env_key": "YAI_SHOW_REASONING", "type": bool},
    "JUSTIFY": {"value": DEFAULT_JUSTIFY, "env_key": "YAI_JUSTIFY", "type": str},
    # Chat history settings
    "CHAT_HISTORY_DIR": {"value": DEFAULT_CHAT_HISTORY_DIR, "env_key": "YAI_CHAT_HISTORY_DIR", "type": str},
    "MAX_SAVED_CHATS": {"value": DEFAULT_MAX_SAVED_CHATS, "env_key": "YAI_MAX_SAVED_CHATS", "type": int},
    # Role settings
    "ROLE_MODIFY_WARNING": {"value": DEFAULT_ROLE_MODIFY_WARNING, "env_key": "YAI_ROLE_MODIFY_WARNING", "type": bool},
    # Function settings
    "ENABLE_FUNCTIONS": {"value": DEFAULT_ENABLE_FUNCTIONS, "env_key": "YAI_ENABLE_FUNCTIONS", "type": bool},
    "SHOW_FUNCTION_OUTPUT": {
        "value": DEFAULT_SHOW_FUNCTION_OUTPUT,
        "env_key": "YAI_SHOW_FUNCTION_OUTPUT",
        "type": bool,
    },
    "ENABLE_MCP": {"value": DEFAULT_ENABLE_MCP, "env_key": "YAI_ENABLE_MCP", "type": bool},
    "SHOW_MCP_OUTPUT": {"value": DEFAULT_SHOW_MCP_OUTPUT, "env_key": "YAI_SHOW_MCP_OUTPUT", "type": bool},
    "MAX_TOOL_CALL_DEPTH": {"value": DEFAULT_MAX_TOOL_CALL_DEPTH, "env_key": "YAI_MAX_TOOL_CALL_DEPTH", "type": int},
    "EXCLUDE_PARAMS": {"value": DEFAULT_EXCLUDE_PARAMS, "env_key": "YAI_EXCLUDE_PARAMS", "type": str},
    # MiniMax specific settings
    "MINIMAX_REASONING_SPLIT": {
        "value": True,
        "env_key": "YAI_MINIMAX_REASONING_SPLIT",
        "type": bool,
    },
}

DEFAULT_CONFIG_INI = f"""[core]
PROVIDER={DEFAULT_PROVIDER}
BASE_URL={DEFAULT_CONFIG_MAP["BASE_URL"]["value"]}
API_KEY={DEFAULT_CONFIG_MAP["API_KEY"]["value"]}
MODEL={DEFAULT_CONFIG_MAP["MODEL"]["value"]}

# auto detect shell and os (or specify manually, e.g., bash, zsh, powershell.exe)
SHELL_NAME={DEFAULT_CONFIG_MAP["SHELL_NAME"]["value"]}
OS_NAME={DEFAULT_CONFIG_MAP["OS_NAME"]["value"]}

DEFAULT_ROLE={DEFAULT_CONFIG_MAP["DEFAULT_ROLE"]["value"]}
# true: streaming response, false: non-streaming
STREAM={DEFAULT_CONFIG_MAP["STREAM"]["value"]}

# LLM parameters
TEMPERATURE={DEFAULT_CONFIG_MAP["TEMPERATURE"]["value"]}
FREQUENCY_PENALTY={DEFAULT_CONFIG_MAP["FREQUENCY_PENALTY"]["value"]}
TOP_P={DEFAULT_CONFIG_MAP["TOP_P"]["value"]}
MAX_TOKENS={DEFAULT_CONFIG_MAP["MAX_TOKENS"]["value"]}
TIMEOUT={DEFAULT_CONFIG_MAP["TIMEOUT"]["value"]}
# json string
EXTRA_HEADERS=
# json string
EXTRA_BODY=
REASONING_EFFORT=

# Interactive mode parameters
INTERACTIVE_ROUND={DEFAULT_CONFIG_MAP["INTERACTIVE_ROUND"]["value"]}

# UI/UX
CODE_THEME={DEFAULT_CONFIG_MAP["CODE_THEME"]["value"]}
# Max entries kept in history file
MAX_HISTORY={DEFAULT_CONFIG_MAP["MAX_HISTORY"]["value"]}
AUTO_SUGGEST={DEFAULT_CONFIG_MAP["AUTO_SUGGEST"]["value"]}
# Print reasoning content or not
SHOW_REASONING={DEFAULT_CONFIG_MAP["SHOW_REASONING"]["value"]}
# Text alignment (default, left, center, right, full)
JUSTIFY={DEFAULT_CONFIG_MAP["JUSTIFY"]["value"]}

# Chat history settings
CHAT_HISTORY_DIR={DEFAULT_CONFIG_MAP["CHAT_HISTORY_DIR"]["value"]}
MAX_SAVED_CHATS={DEFAULT_CONFIG_MAP["MAX_SAVED_CHATS"]["value"]}

# Role settings
# Set to false to disable warnings about modified built-in roles
ROLE_MODIFY_WARNING={DEFAULT_CONFIG_MAP["ROLE_MODIFY_WARNING"]["value"]}

# Function settings
# Set to false to disable sending functions in API requests
ENABLE_FUNCTIONS={DEFAULT_CONFIG_MAP["ENABLE_FUNCTIONS"]["value"]}
# Set to false to disable showing function output when calling functions
SHOW_FUNCTION_OUTPUT={DEFAULT_CONFIG_MAP["SHOW_FUNCTION_OUTPUT"]["value"]}

# MCP settings
# Set to false to disable MCP in API requests
ENABLE_MCP={DEFAULT_CONFIG_MAP["ENABLE_MCP"]["value"]}
# Set to false to disable showing MCP output when calling MCP tools
SHOW_MCP_OUTPUT={DEFAULT_CONFIG_MAP["SHOW_MCP_OUTPUT"]["value"]}

# Maximum number of tool calls to make in a single request
MAX_TOOL_CALL_DEPTH={DEFAULT_CONFIG_MAP["MAX_TOOL_CALL_DEPTH"]["value"]}

# Comma-separated list of API parameters to exclude from requests
# Example: temperature,top_p,frequency_penalty
EXCLUDE_PARAMS=
"""



class YaicliError(Exception):
    """Base exception for yaicli"""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ConfigError(YaicliError):
    """Error in configuration"""


class ConfigMissingError(ConfigError):
    """Missing required key in config"""


class ProviderError(YaicliError):
    """Error in provider"""


class ChatSaveError(YaicliError):
    """Error saving chat"""


class ChatLoadError(YaicliError):
    """Error loading chat"""


class ChatDeleteError(YaicliError):
    """Error deleting chat"""


class MCPToolsError(YaicliError):
    """Error getting MCP tools"""



import configparser
import json
import sys
from dataclasses import dataclass
from functools import lru_cache
from os import getenv
from typing import Any, Optional

from rich import get_console
from rich.console import Console

from .const import (
    CONFIG_PATH,
    DEFAULT_CHAT_HISTORY_DIR,
    DEFAULT_CONFIG_INI,
    DEFAULT_CONFIG_MAP,
    DEFAULT_TEMPERATURE,
    DEFAULT_TOP_P,
)
from .exceptions import ConfigError
from .utils import str2bool


class CasePreservingConfigParser(configparser.RawConfigParser):
    """Case preserving config parser"""

    def optionxform(self, optionstr):
        return optionstr


@dataclass
class ProviderConfig:
    """Provider configuration"""

    api_key: str
    model: str
    base_url: Optional[str] = None
    temperature: float = DEFAULT_TEMPERATURE
    top_p: float = DEFAULT_TOP_P


class Config(dict):
    """Configuration class that loads settings on initialization.

    This class encapsulates the configuration loading logic with priority:
    1. Environment variables (highest priority)
    2. Configuration file
    3. Default values (lowest priority)

    It handles type conversion and validation based on DEFAULT_CONFIG_MAP.
    """

    def __init__(self, console: Optional[Console] = None):
        """Initializes and loads the configuration."""
        self.console = console or get_console()
        super().__init__()
        self.reload()

    def reload(self) -> None:
        """Reload configuration from all sources.

        Follows priority order: env vars > config file > defaults
        """
        # Start with defaults
        self.clear()
        self._load_defaults()

        # Load from config file
        self._load_from_file()

        # Load from environment variables and apply type conversion
        self._load_from_env()
        self._apply_type_conversion()

    def _load_defaults(self) -> None:
        """Load default configuration values as strings."""
        # Direct update instead of creating temporary dict
        for key, config_info in DEFAULT_CONFIG_MAP.items():
            self[key] = config_info["value"]

    def _ensure_version_updated_config_keys(self, config_parser: CasePreservingConfigParser) -> None:
        """Ensure configuration keys added in version updates exist in the config file.

        Uses config parser to check for missing keys instead of full text search.
        Only writes to file if keys are actually missing.
        """
        # Check using config parser instead of full text search
        core_section = config_parser["core"] if config_parser.has_section("core") else {}

        if "CHAT_HISTORY_DIR" not in core_section:
            # Only append if the key is missing
            with open(CONFIG_PATH, "a", encoding="utf-8") as f:
                f.write(f"\nCHAT_HISTORY_DIR={DEFAULT_CHAT_HISTORY_DIR}\n")

    def _load_from_file(self) -> None:
        """Load configuration from the config file.

        Creates default config file if it doesn't exist.
        """
        if not CONFIG_PATH.exists():
            self.console.print("Creating default configuration file.", style="bold yellow", justify=self["JUSTIFY"])
            CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
            with open(CONFIG_PATH, "w", encoding="utf-8") as f:
                f.write(DEFAULT_CONFIG_INI)
            return

        config_parser = CasePreservingConfigParser()
        try:
            config_parser.read(CONFIG_PATH, encoding="utf-8")
        except configparser.DuplicateOptionError as e:
            self.console.print(f"[red]Error:[/red] {e}", justify=self["JUSTIFY"])
            raise ConfigError(str(e)) from None

        # Check if "core" section exists in the config file
        if not config_parser.has_section("core"):
            return

        # Set default values for missing shell/OS info
        core_section = config_parser["core"]
        for key, default_value in {"SHELL_NAME": "Unknown Shell", "OS_NAME": "Unknown OS"}.items():
            if not core_section.get(key, "").strip():
                core_section[key] = default_value

        # Update config with file values
        self.update(dict(core_section))

        # Check if keys added in version updates are missing and add them
        self._ensure_version_updated_config_keys(config_parser)

    def _load_from_env(self) -> None:
        """Load configuration from environment variables.

        Updates the configuration dictionary in-place.
        """
        for key, config_info in DEFAULT_CONFIG_MAP.items():
            env_value = getenv(config_info["env_key"])
            if env_value is not None:
                self[key] = env_value

    def _convert_value(self, raw_value: str, target_type: type, key: str) -> Any:
        """Convert a raw string value to the target type.

        Args:
            raw_value: The raw string value to convert
            target_type: The target type to convert to
            key: The configuration key (for error reporting)

        Returns:
            The converted value or the default value if conversion fails
        """
        if raw_value is None:
            raw_value = DEFAULT_CONFIG_MAP[key]["value"]

        try:
            if target_type is bool:
                return str2bool(raw_value)
            elif target_type in (int, float, str):
                return target_type(raw_value) if raw_value else raw_value
            elif target_type is dict and raw_value:
                return json.loads(raw_value)
            return raw_value
        except (ValueError, TypeError, json.JSONDecodeError) as e:
            # Log warning and fallback to default
            default_value = DEFAULT_CONFIG_MAP[key]["value"]
            self.console.print(
                f"[yellow]Warning:[/] Invalid value '{raw_value}' for '{key}'. "
                f"Expected type '{target_type.__name__}'. Using default value '{default_value}'. Error: {e}",
                style="dim",
                justify=self["JUSTIFY"],
            )

            # Try to convert default value
            try:
                if target_type is bool:
                    return str2bool(default_value)
                elif target_type in (int, float, str):
                    return target_type(default_value)
                elif target_type is dict:
                    return json.loads(default_value)
            except (ValueError, TypeError, json.JSONDecodeError):
                # If default conversion also fails, log error and return raw value
                self.console.print(
                    f"[red]Error:[/red] Could not convert default value for '{key}'. Using raw value.",
                    style="error",
                    justify=self["JUSTIFY"],
                )
                return raw_value

    def _apply_type_conversion(self) -> None:
        """Apply type conversion to configuration values.

        Optimized version that reduces redundant operations and improves error handling.
        """
        for key, config_info in DEFAULT_CONFIG_MAP.items():
            target_type = config_info["type"]
            raw_value = self[key]

            # Skip conversion if already correct type (optimization for common case)
            if isinstance(raw_value, target_type):
                continue

            # Convert the value
            self[key] = self._convert_value(raw_value, target_type, key)


@lru_cache(1)
def get_config() -> Config:
    """Get the configuration singleton"""
    try:
        return Config()
    except ConfigError:
        sys.exit()


cfg = get_config()



import json
from os import getenv
from typing import Any, Dict, Generator, List, Optional, Tuple, cast

from anthropic import Anthropic, AnthropicBedrock, AnthropicVertex, Stream
from anthropic.types import InputJSONDelta, Message, TextDelta
from anthropic.types.raw_message_stream_event import RawMessageStreamEvent
from json_repair import repair_json

from ...config import cfg
from ...console import get_console
from ...exceptions import ConfigMissingError, MCPToolsError
from ...schemas import ChatMessage, LLMResponse, ToolCall
from ..provider import Provider


class AnthropicProvider(Provider):
    """Anthropic provider implementation based on anthropic library"""

    DEFAULT_BASE_URL = "https://api.anthropic.com"
    CLIENT_CLS = Anthropic
    # Base mapping between config keys and API parameter names
    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_tokens": "MAX_TOKENS",
        "timeout": "TIMEOUT",
        "top_k": "TOP_K",
        "stop_sequences": "STOP_SEQUENCES",
        "metadata": "METADATA",
        "extra_body": "EXTRA_BODY",
    }

    def __init__(self, config: dict = cfg, verbose: bool = False, **kwargs):
        self.config = config
        if not self.config.get("API_KEY"):
            raise ValueError("API_KEY is required")
        self.enable_function = self.config["ENABLE_FUNCTIONS"]
        self.enable_mcp = self.config["ENABLE_MCP"]
        self.verbose = verbose

        # Initialize client
        self.client_params = self.get_client_params()
        self.client = self.CLIENT_CLS(**self.client_params)
        self.console = get_console()

    def get_client_params(self) -> Dict[str, Any]:
        """Get the client parameters"""
        # Initialize client params
        client_params = {
            "api_key": self.config["API_KEY"],
            "default_headers": {"X-Title": self.APP_NAME, "HTTP_Referer": self.APP_REFERER},
        }

        # Add base URL if configured
        if self.config.get("BASE_URL"):
            client_params["base_url"] = self.config["BASE_URL"]

        # Add extra headers if set
        if self.config.get("EXTRA_HEADERS"):
            client_params["default_headers"] = {**self.config["EXTRA_HEADERS"], **client_params["default_headers"]}

        # Add timeout if set
        if self.config.get("TIMEOUT"):
            client_params["timeout"] = self.config["TIMEOUT"]

        return client_params

    def get_completion_params_keys(self) -> Dict[str, str]:
        """
        Get the mapping between completion parameter keys and config keys.
        Subclasses can override this method to customize parameter mapping.

        Returns:
            Dict[str, str]: Mapping from API parameter names to config keys
        """
        return self.COMPLETION_PARAMS_KEYS.copy()

    def get_completion_params(self) -> Dict[str, Any]:
        """
        Get the completion parameters based on config and parameter mapping.

        Returns:
            Dict[str, Any]: Parameters for completion API call
        """
        completion_params = {}
        params_keys = self.get_completion_params_keys()
        for api_key, config_key in params_keys.items():
            if self.config.get(config_key, None) is not None and self.config[config_key] != "":
                completion_params[api_key] = self.config[config_key]

        # Apply exclude params filtering
        completion_params = Provider.filter_excluded_params(
            completion_params,
            self.config,
            verbose=self.verbose,
            console=self.console,
        )

        return completion_params

    def completion(
        self,
        messages: List[ChatMessage],
        stream: bool = False,
    ) -> Generator[LLMResponse, None, None]:
        """
        Send completion request to Anthropic and return responses.

        Args:
            messages: List of chat messages to send
            stream: Whether to stream the response

        Yields:
            LLMResponse: Response objects containing content, tool calls, etc.

        Raises:
            ValueError: If messages is empty or invalid
            anthropic.APIError: If API request fails
        """
        anthropic_messages = self._convert_messages(messages)

        params = self.get_completion_params()
        system_prompt, system_prompt_index = self._extract_system_prompt(messages)
        if system_prompt and system_prompt_index is not None:
            anthropic_messages.pop(system_prompt_index)
            params["system"] = system_prompt
        params["messages"] = anthropic_messages
        params["stream"] = stream

        # Add tools if enabled
        tools = []
        if self.enable_function:
            try:
                from ...tools import get_anthropic_schemas

                tools.extend(get_anthropic_schemas())
            except ImportError:
                self.console.print("Function tools not available for Anthropic", style="yellow")

        if self.enable_mcp:
            try:
                from ...tools import get_anthropic_mcp_tools

                mcp_tools = get_anthropic_mcp_tools()
                tools.extend(mcp_tools)
            except (ValueError, FileNotFoundError, MCPToolsError, ImportError) as e:
                self.console.print(f"Failed to load MCP tools: {e}", style="red")

        if tools:
            params["tools"] = tools
            # Disable parallel tool use by default for better compatibility
            disable_parallel = self.config.get("DISABLE_PARALLEL_TOOL_USE", True)
            params["tool_choice"] = {"type": "auto", "disable_parallel_tool_use": disable_parallel}

        if self.verbose:
            self.console.print("System prompt:", params["system"])
            self.console.print("Messages:")
            self.console.print(params["messages"])
            if params.get("tools"):
                self.console.print("Tools:")
                self.console.print(params["tools"])
                self.console.print("Tool choice:", params["tool_choice"])
            if params.get("extra_body"):
                self.console.print("Extra body:")
                self.console.print(params["extra_body"])

        try:
            if stream:
                response = self.client.messages.create(**params)
                yield from self._handle_stream_response(response)
            else:
                response = self.client.messages.create(**params)
                yield from self._handle_normal_response(response)
        except Exception as e:
            self.console.print(f"Error: {e}", style="red")
            raise

    def _extract_system_prompt(self, messages: List[ChatMessage]) -> Tuple[Optional[str], Optional[int]]:
        """Extract system prompt from messages"""
        for index, message in enumerate(messages):
            if message.role == "system":
                return message.content, index
        return None, None

    def _handle_normal_response(self, response: Message) -> Generator[LLMResponse, None, None]:
        """Handle normal (non-streaming) response"""
        if not response.content:
            yield LLMResponse(content=json.dumps(response.model_dump()), finish_reason="stop")
            return

        # Extract content from all blocks in a single pass
        text_content = ""
        thinking_content = ""
        tool_call: Optional[ToolCall] = None

        for block in response.content:
            if block.type == "text" and hasattr(block, "text"):
                text_content += block.text
            elif block.type == "thinking":
                # Handle thinking blocks
                thinking_content += getattr(block, "thinking", "")
            elif block.type == "tool_use":
                # Handle tool use blocks
                tool_call = ToolCall(
                    id=getattr(block, "id", ""),
                    name=block.name,
                    # String input is already valid JSON, no need to convert
                    arguments=block.input if isinstance(block.input, str) else json.dumps(block.input),
                )

        # Get stop reason
        finish_reason = response.stop_reason or "stop"

        # Yield response with all content types
        yield LLMResponse(
            reasoning=thinking_content if thinking_content else None,
            content=text_content,
            finish_reason=finish_reason,
            tool_call=tool_call,
        )

    def _handle_stream_response(self, response: Stream[RawMessageStreamEvent]) -> Generator[LLMResponse, None, None]:
        """Handle streaming response from Anthropic API"""
        tool_call: Optional[ToolCall] = None
        tool_call_id = ""
        tool_call_name = ""
        tool_call_input = ""

        # Process each chunk in the response stream
        for chunk in response:
            # Handle different event types
            if chunk.type == "message_start":
                # Message start, no need to handle
                continue

            elif chunk.type == "content_block_start":
                # Content block start
                if hasattr(chunk, "content_block") and getattr(chunk.content_block, "type", "") == "tool_use":
                    # Tool call block - initialize tool call data
                    tool_data = chunk.content_block
                    tool_call_id = getattr(tool_data, "id", "")
                    tool_call_name = getattr(tool_data, "name", "")
                    tool_call_input = ""  # Will be built incrementally

            elif chunk.type == "content_block_delta":
                # Content delta update
                if hasattr(chunk, "delta"):
                    delta = chunk.delta

                    # Handle text delta
                    if hasattr(delta, "text"):
                        delta = cast(TextDelta, delta)
                        content = delta.text or ""
                        if content:
                            yield LLMResponse(content=content, tool_call=None)

                    # Handle tool call input delta
                    elif hasattr(delta, "partial_json"):
                        delta = cast(InputJSONDelta, delta)
                        tool_call_input += delta.partial_json or ""

            elif chunk.type == "content_block_stop":
                # Content block stop - finalize tool call if it exists
                if tool_call_id and tool_call_name:
                    tool_call = ToolCall(
                        id=tool_call_id,
                        name=tool_call_name,
                        arguments=repair_json(tool_call_input),
                    )
                    # Reset tool call data
                    tool_call_id = ""
                    tool_call_name = ""
                    tool_call_input = ""

            elif chunk.type == "message_delta":
                # Message delta update - contains stop reason
                if hasattr(chunk, "delta") and hasattr(chunk.delta, "stop_reason"):
                    finish_reason = chunk.delta.stop_reason
                    if finish_reason:
                        yield LLMResponse(content="", finish_reason=finish_reason, tool_call=tool_call)

            elif chunk.type == "message_stop":
                # Message stop - final response
                yield LLMResponse(content="", finish_reason="stop", tool_call=tool_call)

    def detect_tool_role(self) -> str:
        """Return the role that should be used for tool responses"""
        return "tool"

    def _convert_messages(self, messages: List[ChatMessage]) -> List[Dict[str, Any]]:
        """Convert ChatMessage list to Anthropic format with proper tool result handling"""
        converted_messages = []
        i = 0

        while i < len(messages):
            msg = messages[i]

            # Handle regular messages
            if msg.role != "tool":
                message: Dict[str, Any] = {"role": msg.role, "content": msg.content or ""}

                # Handle tool calls in assistant messages
                if msg.role == "assistant" and msg.tool_calls:
                    content = []
                    if msg.content:
                        content.append({"type": "text", "text": msg.content})

                    for tool_call in msg.tool_calls:
                        content.append(
                            {
                                "type": "tool_use",
                                "id": tool_call.id,
                                "name": tool_call.name,
                                "input": json.loads(tool_call.arguments),
                            }
                        )

                    message["content"] = content

                converted_messages.append(message)
                i += 1
            else:
                # Handle tool results - collect consecutive tool messages
                tool_results = []
                while i < len(messages) and messages[i].role == "tool":
                    tool_msg = messages[i]
                    tool_results.append(
                        {"type": "tool_result", "tool_use_id": tool_msg.tool_call_id, "content": tool_msg.content or ""}
                    )
                    i += 1

                # Add tool results as a user message
                if tool_results:
                    converted_messages.append({"role": "user", "content": tool_results})

        return converted_messages


class AnthropicBedrockProvider(AnthropicProvider):
    """Anthropic Bedrock provider implementation based on anthropic library"""

    CLIENT_CLS = AnthropicBedrock
    CLIENT_PARAMS_KEY_ENV_MAP = {
        "AWS_SECRET_ACCESS_KEY": "AWS_SECRET_ACCESS_KEY",
        "AWS_ACCESS_KEY_ID": "AWS_ACCESS_KEY_ID",
        "AWS_SESSION_TOKEN": "AWS_SESSION_TOKEN",
        "AWS_REGION": "AWS_REGION",
    }

    def get_client_params(self) -> Dict[str, Any]:
        """Get the client parameters
        client = AnthropicBedrock(
            # Authenticate by either providing the keys below or use the default AWS credential providers, such as
            # using ~/.aws/credentials or the "AWS_SECRET_ACCESS_KEY" and "AWS_ACCESS_KEY_ID" environment variables.
            aws_access_key="<access key>",
            aws_secret_key="<secret key>",
            # Temporary credentials can be used with aws_session_token.
            # Read more at https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_temp.html.
            aws_session_token="<session_token>",
            # aws_region changes the aws region to which the request is made. By default, we read AWS_REGION,
            # and if that's not present, we default to us-east-1. Note that we do not read ~/.aws/config for the region.
            aws_region="us-west-2",
        )
        """
        client_params = super().get_client_params()
        for config_key, env_key in self.CLIENT_PARAMS_KEY_ENV_MAP.items():
            if not self.config.get(config_key):
                v = getenv(env_key)
                if not v:
                    raise ConfigMissingError(
                        f"{env_key} is required. Please set `{config_key}` in config or environment variable `{env_key}`"
                    )
                self.config[config_key] = v

        client_params["aws_access_key"] = self.config["AWS_ACCESS_KEY_ID"]
        client_params["aws_secret_key"] = self.config["AWS_SECRET_ACCESS_KEY"]
        client_params["aws_session_token"] = self.config["AWS_SESSION_TOKEN"]
        client_params["aws_region"] = self.config["AWS_REGION"]
        return client_params


class AnthropicVertexProvider(AnthropicProvider):
    """Anthropic Vertex provider implementation based on anthropic library"""

    CLIENT_CLS = AnthropicVertex
    CLIENT_PARAMS_KEY_ENV_MAP = {
        "PROJECT_ID": "PROJECT_ID",
        "CLOUD_ML_REGION": "CLOUD_ML_REGION",
    }

    def get_client_params(self) -> Dict[str, Any]:
        """Get the client parameters"""
        client_params = super().get_client_params()
        for config_key, env_key in self.CLIENT_PARAMS_KEY_ENV_MAP.items():
            if not self.config.get(config_key):
                v = getenv(env_key)
                if not v:
                    raise ConfigMissingError(
                        f"{env_key} is required. Please set `{config_key}` in config or environment variable `{env_key}`"
                    )
                self.config[config_key] = v
        client_params["project_id"] = self.config["PROJECT_ID"]
        client_params["region"] = self.config["CLOUD_ML_REGION"]
        return client_params



import json
import re
import uuid
from typing import Any, Dict, Generator, Optional

from ...schemas import LLMResponse, ToolCall
from .anthropic_provider import AnthropicProvider
from .openai_provider import OpenAIProvider


def _clean_longcat_reasoning(reasoning: str) -> str:
    """
    Remove LongCat tool call tags from reasoning/content.

    Args:
        reasoning: The reasoning/content that may contain a tool call

    Returns:
        Cleaned reasoning/content without tool call tags
    """
    if not reasoning:
        return reasoning

    # Pattern to match and remove the LongCat tool call format
    pattern = r"<longcat_tool_call>.*?</longcat_tool_call>"
    return re.sub(pattern, "", reasoning, flags=re.DOTALL).strip()


def _parse_and_clean_longcat_content(
    content: str, response_id: Optional[str] = None, verbose: bool = False, console=None
) -> tuple[Optional[ToolCall], str]:
    """
    Parse LongCat-specific tool call format and return both the tool call and cleaned content.

    The format is:
    <longcat_tool_call>function_name
    <longcat_arg_key>key1</longcat_arg_key>
    <longcat_arg_value>value1</longcat_arg_value>
    <longcat_arg_key>key2</longcat_arg_key>
    <longcat_arg_value>value2</longcat_arg_value>
    ...
    </longcat_tool_call>

    Args:
        content: The reasoning/content that may contain a tool call
        response_id: Optional response ID to use for tool call ID
        verbose: Whether to print verbose logging
        console: Console object for verbose output

    Returns:
        Tuple of (ToolCall or None, cleaned_content)
    """
    if not content:
        return None, content

    # Pattern to match the LongCat tool call format with multiple arguments
    pattern = r"<longcat_tool_call>(\w+)(.*?)</longcat_tool_call>"

    match = re.search(pattern, content, re.DOTALL)
    if match:
        function_name = match.group(1)
        args_section = match.group(2)

        # Parse all key-value pairs
        args_pattern = r"<longcat_arg_key>(\w+)</longcat_arg_key>\s*<longcat_arg_value>(.*?)</longcat_arg_value>"
        args_matches = re.findall(args_pattern, args_section, re.DOTALL)

        if args_matches:
            # Build arguments dictionary
            arguments_dict = {}
            for arg_key, arg_value in args_matches:
                arguments_dict[arg_key] = arg_value.strip()

            # Print verbose message
            if verbose and console:
                console.print(f"[LongCat] Extracted tool call: {function_name}")

            # Build JSON arguments string
            arguments = json.dumps(arguments_dict)

            # Use response.id if available, otherwise generate UUID
            tool_call_id = response_id if response_id else f"longcat_{uuid.uuid4().hex[:8]}"

            tool_call = ToolCall(id=tool_call_id, name=function_name, arguments=arguments)
            cleaned_content = _clean_longcat_reasoning(content)

            return tool_call, cleaned_content

    return None, content


class LongCatProvider(OpenAIProvider):
    """LongCat provider implementation based on openai-compatible API"""

    DEFAULT_BASE_URL = "https://api.longcat.chat/openai"
    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_tokens": "MAX_TOKENS",
        "timeout": "TIMEOUT",
        "extra_body": "EXTRA_BODY",
        "frequency_penalty": "FREQUENCY_PENALTY",
    }

    def _handle_normal_response(self, response) -> Generator[LLMResponse, None, None]:
        """Handle normal (non-streaming) response with LongCat tool call parsing"""
        # First get the standard response handling
        for std_response in super()._handle_normal_response(response):
            # Check if there's a tool call hidden in reasoning when finish_reason is "stop"
            if std_response.finish_reason == "stop" and std_response.reasoning:
                # Get response.id if available
                response_id = getattr(response, "id", None)
                tool_call, cleaned_reasoning = _parse_and_clean_longcat_content(
                    std_response.reasoning, response_id, self.verbose, self.console
                )
                if tool_call:
                    yield LLMResponse(
                        reasoning=cleaned_reasoning,
                        content=std_response.content,
                        finish_reason="tool_calls",
                        tool_call=tool_call,
                    )
                else:
                    yield std_response
            else:
                yield std_response

    def _handle_stream_response(self, response) -> Generator[LLMResponse, None, None]:
        """Handle streaming response with LongCat tool call parsing"""
        # Accumulate reasoning content to check for tool calls
        accumulated_reasoning = ""
        tool_call = None
        cleaned_reasoning = ""
        response_id = getattr(response, "id", None)

        for chunk_response in super()._handle_stream_response(response):
            # Accumulate reasoning content
            if chunk_response.reasoning:
                accumulated_reasoning += chunk_response.reasoning

            # If we haven't found a tool call yet and we're accumulating reasoning
            if tool_call is None and accumulated_reasoning:
                tool_call, cleaned_reasoning = _parse_and_clean_longcat_content(
                    accumulated_reasoning, response_id, self.verbose, self.console
                )

            # If finish_reason is "stop" and we found a tool call in reasoning
            if chunk_response.finish_reason == "stop" and tool_call:
                yield LLMResponse(
                    reasoning=cleaned_reasoning,
                    content=chunk_response.content,
                    finish_reason="tool_calls",
                    tool_call=tool_call,
                )
            # If finish_reason is "tool_calls" from standard handling
            elif chunk_response.finish_reason == "tool_calls":
                yield chunk_response
            # Otherwise yield as-is
            else:
                yield LLMResponse(
                    reasoning=chunk_response.reasoning,
                    content=chunk_response.content,
                    finish_reason=chunk_response.finish_reason,
                    tool_call=None,
                )


class LongCatAnthropicProvider(AnthropicProvider):
    """LongCat provider implementation based on Anthropic-compatible API"""

    DEFAULT_BASE_URL = "https://api.longcat.chat/anthropic"
    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_tokens": "MAX_TOKENS",
        "timeout": "TIMEOUT",
        "top_k": "TOP_K",
        "stop_sequences": "STOP_SEQUENCES",
        "metadata": "METADATA",
        "extra_body": "EXTRA_BODY",
    }

    def get_client_params(self) -> Dict[str, Any]:
        """Get the client parameters with Bearer token in Authorization header"""
        client_params = {
            "api_key": self.config["API_KEY"],
            "base_url": self.config.get("BASE_URL") or self.DEFAULT_BASE_URL,
            "default_headers": {
                "Authorization": f"Bearer {self.config['API_KEY']}",
                "X-Title": self.APP_NAME,
                "HTTP_Referer": self.APP_REFERER,
            },
        }

        # Add extra headers if set
        if self.config.get("EXTRA_HEADERS"):
            client_params["default_headers"] = {
                **self.config["EXTRA_HEADERS"],
                **client_params["default_headers"],
            }

        # Add timeout if set
        if self.config.get("TIMEOUT"):
            client_params["timeout"] = self.config["TIMEOUT"]

        return client_params

    def _handle_normal_response(self, response) -> Generator[LLMResponse, None, None]:
        """Handle normal (non-streaming) response with LongCat tool call parsing"""
        # First get the standard response handling
        for std_response in super()._handle_normal_response(response):
            # If parent already found tool_calls, yield as-is
            if std_response.finish_reason == "tool_use" and std_response.tool_call:
                yield std_response
            # Check if there's a tool call hidden in reasoning when finish_reason is "end_turn"
            elif std_response.finish_reason == "end_turn":
                # Get response.id if available
                response_id = getattr(response, "id", None)

                # Check reasoning field first (for OpenAI-compatible)
                if std_response.reasoning:
                    tool_call, cleaned_reasoning = _parse_and_clean_longcat_content(
                        std_response.reasoning, response_id, self.verbose, self.console
                    )
                    if tool_call:
                        yield LLMResponse(
                            reasoning=cleaned_reasoning,
                            content=std_response.content,
                            finish_reason="tool_use",
                            tool_call=tool_call,
                        )
                        return

                # Check content field (for Anthropic-compatible)
                if std_response.content:
                    tool_call, cleaned_content = _parse_and_clean_longcat_content(
                        std_response.content, response_id, self.verbose, self.console
                    )
                    if tool_call:
                        yield LLMResponse(
                            reasoning=std_response.reasoning,
                            content=cleaned_content,
                            finish_reason="tool_use",
                            tool_call=tool_call,
                        )
                        return

                # No tool call found, yield original response
                yield std_response
            else:
                yield std_response

    def _handle_stream_response(self, response) -> Generator[LLMResponse, None, None]:
        """Handle streaming response with LongCat tool call parsing"""
        # Accumulate content to check for tool calls
        accumulated_content = ""
        tool_call = None
        cleaned_content = ""
        response_id = getattr(response, "id", None)

        for chunk_response in super()._handle_stream_response(response):
            # Accumulate content
            if chunk_response.content:
                accumulated_content += chunk_response.content

            # If we haven't found a tool call yet and we're accumulating content
            if tool_call is None and accumulated_content:
                tool_call, cleaned_content = _parse_and_clean_longcat_content(
                    accumulated_content, response_id, self.verbose, self.console
                )

            # If finish_reason is "stop" and we found a tool call in content
            if chunk_response.finish_reason == "stop" and tool_call:
                yield LLMResponse(
                    content=cleaned_content,
                    finish_reason="tool_use",
                    tool_call=tool_call,
                )
            # If finish_reason is "tool_calls" from standard handling
            elif chunk_response.finish_reason == "tool_use":
                yield chunk_response
            # Otherwise yield as-is
            else:
                yield LLMResponse(
                    content=chunk_response.content,
                    finish_reason=chunk_response.finish_reason,
                    tool_call=None,
                )



import subprocess
import sys
import time
import traceback
from os import devnull
from pathlib import Path
from typing import Optional, Union

import typer
from prompt_toolkit import PromptSession, prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Prompt

from .chat import Chat, FileChatManager, chat_mgr
from .cmd_handler import CmdHandler
from .completer import AtPathCompleter
from .config import cfg
from .console import get_console
from .const import (
    CHAT_MODE,
    CMD_CLEAR,
    CMD_DELETE_CHAT,
    CMD_EXIT,
    CMD_HELP,
    CMD_HISTORY,
    CMD_LIST_CHATS,
    CMD_LOAD_CHAT,
    CMD_MODE,
    CMD_SAVE_CHAT,
    CONFIG_PATH,
    DEFAULT_OS_NAME,
    DEFAULT_SHELL_NAME,
    EXEC_MODE,
    HISTORY_FILE,
    TEMP_MODE,
    DefaultRoleNames,
)
from .context import ContextManager, ctx_mgr
from .exceptions import ChatSaveError, YaicliError
from .history import LimitedFileHistory
from .llms import LLMClient
from .printer import Printer
from .role import Role, RoleManager, role_mgr
from .schemas import ChatMessage
from .utils import detect_os, detect_shell, filter_command


class CLI:
    __slots__ = (
        "verbose",
        "init_role",
        "role_name",
        "console",
        "chat_manager",
        "role_manager",
        "role",
        "printer",
        "client",
        "cmd_handler",
        "bindings",
        "current_mode",
        "interactive_round",
        "chat_start_time",
        "is_temp_session",
        "chat",
        "chat_history_dir",
        "session",
        "history",
        "context_manager",
    )

    def __init__(
        self,
        verbose: bool = False,
        role: str = DefaultRoleNames.DEFAULT,
        chat_manager: Optional[FileChatManager] = None,
        role_manager: Optional[RoleManager] = None,
        context_manager: Optional[ContextManager] = None,
        client=None,
    ):
        self.verbose: bool = verbose
        # --role can specify a role when enter interactive chat
        # TAB will switch between role and shell
        self.init_role: str = role
        self.role_name: str = role

        self.console = get_console()
        self.chat_manager = chat_manager or chat_mgr
        self.role_manager = role_manager or role_mgr
        self.context_manager = context_manager or ctx_mgr
        self.role: Role = self.role_manager.get_role(self.role_name)
        self.printer = Printer()
        self.client = client or self._create_client()
        self.cmd_handler = CmdHandler(self)

        self.bindings = KeyBindings()

        self.current_mode: str = TEMP_MODE

        self.interactive_round = cfg["INTERACTIVE_ROUND"]
        self.chat_start_time = None
        self.is_temp_session = True
        self.chat = Chat(title="", history=[])

        # Get and create chat history directory from configuration
        self.chat_history_dir = Path(cfg["CHAT_HISTORY_DIR"])
        # if not self.chat_history_dir.exists():
        #     self.chat_history_dir.mkdir(parents=True, exist_ok=True)

        # Detect OS and Shell if set to auto
        if cfg["OS_NAME"] == DEFAULT_OS_NAME:
            cfg["OS_NAME"] = detect_os(cfg)
        if cfg["SHELL_NAME"] == DEFAULT_SHELL_NAME:
            cfg["SHELL_NAME"] = detect_shell(cfg)

        if self.verbose:
            # Print verbose configuration
            self.console.print("Loading Configuration:", style="bold cyan")
            self.console.print(f"Config file path: {CONFIG_PATH}")
            for key, value in cfg.items():
                display_value = "****" if key == "API_KEY" and value else value
                self.console.print(f"  {key:<20}: {display_value}")
            self.console.print(f"Current role: {self.role_name}")
            self.console.print(Markdown("---", code_theme=cfg["CODE_THEME"]))

        # Disable prompt_toolkit warning when use non-tty input,
        # e.g. when use pipe or redirect
        _origin_stderr = None
        if not sys.stdin.isatty():
            _origin_stderr = sys.stderr
            sys.stderr = open(devnull, "w", encoding="utf-8")
        try:
            self.session = PromptSession(key_bindings=self.bindings)
        finally:
            if _origin_stderr:
                sys.stderr.flush()
                sys.stderr.close()
                sys.stderr = _origin_stderr

    def set_role(self, role_name: str) -> None:
        self.role_name = role_name
        self.role = self.role_manager.get_role(role_name)
        if role_name in (DefaultRoleNames.CODER, DefaultRoleNames.SHELL):
            cfg["ENABLE_FUNCTIONS"] = False
        if role_name == DefaultRoleNames.CODER:
            self.printer = Printer(content_markdown=False)
        elif role_name == DefaultRoleNames.SHELL:
            self.current_mode = EXEC_MODE

    @classmethod
    def evaluate_role_name(cls, code: bool = False, shell: bool = False, role: str = ""):
        """
        Judge the role based on the code, shell, and role options.
        Code and shell are highest priority, then role, then default.
        """
        if code is True:
            return DefaultRoleNames.CODER
        if shell is True:
            return DefaultRoleNames.SHELL
        if role:
            return role
        return DefaultRoleNames.DEFAULT

    def get_prompt_tokens(self) -> list[tuple[str, str]]:
        """Return prompt tokens for current mode"""
        mode_icon = "💬" if self.current_mode == CHAT_MODE else "🚀" if self.current_mode == EXEC_MODE else "📝"
        return [("class:qmark", f" {mode_icon} "), ("class:prompt", "> ")]

    def _check_history_len(self) -> None:
        """Check history length and remove the oldest messages if necessary"""
        target_len = self.interactive_round * 2
        if len(self.chat.history) > target_len:
            self.chat.history = self.chat.history[-target_len:]
            if self.verbose:
                self.console.print(f"Dialogue trimmed to {target_len} messages.", style="dim")

    # ------------------- Chat Command Methods -------------------
    def _save_chat(self, title: Union[str, None] = None) -> None:
        """Save current chat history to a file using session manager."""
        # Update title if provided
        if title:
            self.chat.title = title

        # Save chat and get the saved title back
        try:
            saved_title = self.chat_manager.save_chat(self.chat)
        except ChatSaveError as e:
            self.console.print(f"Failed to save chat: {e}", style="red")
            return

        # Session list will be refreshed automatically by the save method
        self.console.print(f"Chat saved as: {saved_title}", style="bold green")

        # Mark session as persistent if it was temporary
        if self.is_temp_session:
            self.is_temp_session = False
            self.chat_start_time = int(time.time())
            self.console.print(
                "Session is now marked as persistent and will be auto-saved on exit.", style="bold green"
            )

    def _list_chats(self) -> None:
        """List all saved chat sessions using session manager."""
        chats: list[Chat] = self.chat_manager.list_chats()

        if not chats:
            self.console.print("No saved chats found.", style="yellow")
            return

        self.console.print("Saved Chats:", style="bold underline")
        for chat in chats:
            index = chat.idx
            title = chat.title
            date = chat.date

            if date:
                self.console.print(f"[dim]{index}.[/dim] [bold blue]{title}[/bold blue] - {date}")
            else:
                self.console.print(f"[dim]{index}.[/dim] [bold blue]{title}[/bold blue]")

    def _refresh_chats(self) -> None:
        """Force refresh the chat list."""
        self.chat_manager.refresh_chats()

    def _load_chat_by_index(self, index: str) -> bool:
        """Load a chat session by its index using session manager."""
        if not self.chat_manager.validate_chat_index(index):
            self.console.print("Invalid chat index.", style="bold red")
            return False

        chat_data = self.chat_manager.load_chat_by_index(index)

        if not chat_data:
            self.console.print("Invalid chat index or chat not found.", style="bold red")
            return False

        self.chat = chat_data
        self.chat_start_time = chat_data.date
        self.is_temp_session = False

        self.console.print(f"Loaded chat: {self.chat.title}", style="bold green")
        return True

    def _delete_chat_by_index(self, index: str) -> bool:
        """Delete a chat session by its index using session manager."""
        if not self.chat_manager.validate_chat_index(index):
            self.console.print("Invalid chat index.", style="bold red")
            return False

        chat_data = self.chat_manager.load_chat_by_index(index)

        if not chat_data:
            self.console.print("Invalid chat index or chat not found.", style="bold red")
            return False

        if chat_data.path is None:
            self.console.print(f"Chat has no associated file to delete: {chat_data.title}", style="bold red")
            return False

        if self.chat_manager.delete_chat(chat_data.path):
            self.console.print(f"Deleted chat: {chat_data.title}", style="bold green")
            return True
        else:
            self.console.print(f"Failed to delete chat: {chat_data.title}", style="bold red")
            return False

    # ------------------- Special commands -------------------
    def _handle_special_commands(self, user_input: str) -> Union[bool, str]:
        """Handle special command return: True-continue loop, False-exit loop, str-non-special command"""
        return self.cmd_handler.handle_command(user_input)

    def _build_messages(self, user_input: str) -> list[ChatMessage]:
        """Build message list for LLM API with @ file references expanded."""
        # Create the message list with system prompt
        messages = [ChatMessage(role="system", content=self.role.prompt)]

        # Add context messages if any
        context_msgs = self.context_manager.get_context_messages()
        if context_msgs:
            messages.extend(context_msgs)

        # Parse and add temporary @ file references
        at_refs_content, cleaned_input = self.context_manager.parse_at_references(user_input)
        if at_refs_content:
            messages.append(ChatMessage(role="system", content=at_refs_content))

        # Add previous conversation if available
        for msg in self.chat.history:
            messages.append(msg)

        # Add user input (with @ references cleaned up)
        messages.append(ChatMessage(role="user", content=cleaned_input))
        return messages

    def _handle_llm_response(self, user_input: str) -> tuple[Optional[str], list[ChatMessage]]:
        """Get response from API (streaming or normal) and print it.
        Returns the full content string or None if an error occurred.

        Args:
            user_input (str): The user's input text.

        Returns:
            Optional[str]: The assistant's response content or None if an error occurred.
            list[ChatMessage]: The updated message history.
        """
        messages = self._build_messages(user_input)
        if self.role.name != DefaultRoleNames.CODER:
            self.console.print("Assistant:", style="bold green")
        try:
            response_iterator = self.client.completion_with_tools(messages, stream=cfg["STREAM"])

            content, _ = self.printer.display_stream(response_iterator)

            # The 'messages' list is modified by the client in-place
            return content, messages
        except Exception as e:
            self.console.print(f"Error processing LLM response: {e}", style="red")
            if self.verbose:
                traceback.print_exc()
            return None, messages

    def _process_user_input(self, user_input: str) -> bool:
        """Process user input: get response, print, update history, maybe execute.
        Returns True to continue REPL, False to exit on critical error.
        """
        content, updated_messages = self._handle_llm_response(user_input)

        if content is None and not any(msg.tool_calls for msg in updated_messages):
            return True

        # The client modifies the message list in place, so the updated_messages
        # contains the full history of the turn (system, history, user, assistant, tools).
        # We replace the old history with the new one, removing system messages (role prompt, context, etc.)
        # to ensure they remain temporary and don't pollute the conversation history.
        if updated_messages:
            self.chat.history = [msg for msg in updated_messages if msg.role != "system"]

        self._check_history_len()

        if self.current_mode == EXEC_MODE:
            self._confirm_and_execute(content or "")
        return True

    def _confirm_and_execute(self, raw_content: str) -> None:
        """Review, edit and execute the command"""
        cmd = filter_command(raw_content)
        if not cmd:
            self.console.print("No command generated or command is empty.", style="bold red")
            return
        self.console.print(
            Panel(cmd, title="Suggest Command", title_align="left", border_style="bold magenta", expand=False)
        )
        _input = Prompt.ask(
            r"Execute command? \[e]dit, \[y]es, \[n]o",
            choices=["y", "n", "e"],
            default="y",
            case_sensitive=False,
            show_choices=False,
        )
        executed_cmd = None
        if _input == "y":
            executed_cmd = cmd
        elif _input == "e":
            try:
                edited_cmd = prompt("Edit command: ", default=cmd).strip()
                if edited_cmd and edited_cmd != cmd:
                    executed_cmd = edited_cmd
                elif edited_cmd:
                    executed_cmd = cmd
                else:
                    self.console.print("Execution cancelled.", style="yellow")
            except EOFError:
                self.console.print("\nEdit cancelled.", style="yellow")
                return
        if executed_cmd:
            self.console.print("Executing...", style="bold green")
            try:
                subprocess.call(executed_cmd, shell=True)
            except Exception as e:
                self.console.print(f"Failed to execute command: {e}", style="red")
        elif _input != "e":
            self.console.print("Execution cancelled.", style="yellow")

    # ------------------- REPL Methods -------------------
    def prepare_chat_loop(self) -> None:
        """Setup key bindings and history for interactive modes."""
        self.current_mode = CHAT_MODE
        self._setup_key_bindings()
        HISTORY_FILE.touch(exist_ok=True)

        # Set up the prompt session with command history
        try:
            self.session = PromptSession(
                key_bindings=self.bindings,
                history=LimitedFileHistory(HISTORY_FILE, max_entries=self.interactive_round),
                auto_suggest=AutoSuggestFromHistory() if cfg.get("AUTO_SUGGEST", True) else None,
                enable_history_search=True,
                completer=AtPathCompleter(),
                complete_while_typing=True,  # Enable auto-completion for @ trigger
                complete_in_thread=True,  # Don't block on completion
            )
        except Exception as e:
            self.console.print(f"Error initializing prompt session history: {e}", style="red")
            self.session = PromptSession(key_bindings=self.bindings)

    def _setup_key_bindings(self) -> None:
        """Setup keyboard shortcuts with Shift+Tab for mode switching."""

        @self.bindings.add("s-tab")  # Shift+Tab to switch mode
        def _(event: KeyPressEvent) -> None:
            """Switch between chat and exec mode."""
            self.current_mode = EXEC_MODE if self.current_mode == CHAT_MODE else CHAT_MODE
            self.set_role(DefaultRoleNames.SHELL if self.current_mode == EXEC_MODE else self.init_role)

    def _print_welcome_message(self) -> None:
        """Prints the initial welcome banner and instructions."""
        self.console.print(
            """
 ██    ██  █████  ██  ██████ ██      ██
  ██  ██  ██   ██ ██ ██      ██      ██
   ████   ███████ ██ ██      ██      ██
    ██    ██   ██ ██ ██      ██      ██
    ██    ██   ██ ██  ██████ ███████ ██
""",
            style="bold cyan",
        )
        self.console.print("Welcome to YAICLI!", style="bold")

        # Display session type
        if self.is_temp_session:
            self.console.print("Current: [bold yellow]Temporary Session[/bold yellow] (use /save to make persistent)")
        else:
            self.console.print(
                f"Current: [bold green]Persistent Session[/bold green]{f': {self.chat.title}' if self.chat.title else ''}"
            )
        self.print_help()

    def print_help(self):
        self.console.print("Press [bold yellow]Shift+Tab[/bold yellow] to switch mode")
        help_cmd = "|".join(CMD_HELP)
        self.console.print(f"{help_cmd:<19}: Show help message")
        self.console.print(f"{CMD_CLEAR:<19}: Clear chat history")
        self.console.print(f"{CMD_HISTORY:<19}: Show chat history")
        self.console.print(f"{CMD_LIST_CHATS:<19}: List saved chats")
        save_cmd = f"{CMD_SAVE_CHAT} <title>"
        self.console.print(f"{save_cmd:<19}: Save current chat")
        load_cmd = f"{CMD_LOAD_CHAT} <index>"
        self.console.print(f"{load_cmd:<19}: Load a saved chat")
        delete_cmd = f"{CMD_DELETE_CHAT} <index>"
        self.console.print(f"{delete_cmd:<19}: Delete a saved chat")
        self.console.print(f"{'/add <path>':<19}: Add @file/dir to context")
        self.console.print(f"{'/context, /ctx':<19}: Manage context (list, add, remove, clear)")
        self.console.print("[dim]  Tip: Type '@' for path completion, use Tab/arrows to select[/dim]")
        self.console.print(f"{'!<command>':<19}: Execute shell command directly (e.g., !ls -al)")
        cmd_exit = f"{CMD_EXIT}|Ctrl+D|Ctrl+C"
        self.console.print(f"{cmd_exit:<19}: Exit")
        cmd_mode = f"{CMD_MODE} {CHAT_MODE}|{EXEC_MODE}"
        self.console.print(f"{cmd_mode:<19}: Switch mode (Case insensitive)", style="dim")

    def _run_repl(self) -> None:
        """Run the main Read-Eval-Print Loop (REPL)."""
        self.prepare_chat_loop()
        self._print_welcome_message()

        # Main REPL loop
        while True:
            self.console.print(Markdown("---", code_theme=cfg["CODE_THEME"]))
            try:
                # Get user input
                user_input = self.session.prompt(self.get_prompt_tokens)
                user_input = user_input.strip()
                if not user_input:
                    continue

                # Handle special commands
                _continue = self._handle_special_commands(user_input)
                if _continue is False:  # Exit command
                    break
                if _continue is True:  # Other special command
                    continue

                # Process regular chat input
                try:
                    if not self._process_user_input(user_input):
                        break
                except KeyboardInterrupt:
                    self.console.print("KeyboardInterrupt", style="yellow")
                    continue
            except (KeyboardInterrupt, EOFError):
                break

        # Auto-save chat history when exiting if there are messages and not a temporary session
        if not self.is_temp_session and self.chat.history:
            self._save_chat(self.chat.title)

        self.console.print("\nExiting YAICLI... Goodbye!", style="bold green")

    def _run_once(self, user_input: str, shell: bool = False, code: bool = False) -> None:
        """Handle default mode"""
        self.set_role(self.evaluate_role_name(code, shell, self.init_role))
        self._process_user_input(user_input)

    def run(self, chat: bool = False, shell: bool = False, code: bool = False, user_input: Optional[str] = None):
        if not user_input and not chat:
            self.console.print("No input provided.", style="bold red")
            raise typer.Abort()

        if chat:
            # If user provided a title, try to load that chat
            if user_input and isinstance(user_input, str):
                loaded_chat = self.chat_manager.load_chat_by_title(user_input)
                if loaded_chat:
                    self.chat = loaded_chat
                    self.is_temp_session = False
            # Run the interactive chat REPL
            self._run_repl()
        else:
            # Run in single-use mode
            self._run_once(user_input or "", shell=shell, code=code)

    def _create_client(self):
        """Create an LLM client instance based on configuration"""
        try:
            return LLMClient(provider_name=cfg["PROVIDER"].lower(), verbose=self.verbose, config=cfg)
        except YaicliError as e:
            self.console.print(f"Error creating client: {e}", style="red")
            raise typer.Abort()



import json

import httpx
from yaicli.function_schema import OpenAISchema
from pydantic import Field


class Function(OpenAISchema):
    """Query the weather from wttr.in for today and the next two days by city name in English."""

    city: str = Field(description="The city to query the weather of.")
    language: str = Field(
        description="The language of the weather information.",
        default="en",
        examples=["en", "zh", "ja", "ko", "fr", "de", "es", "it"],
    )

    class Config:
        title = "get_weather"

    @classmethod
    def execute(cls, city: str, language: str = "en"):
        """execute the function"""
        url = f"https://wttr.in/{city}"
        params = {
            "format": "j1",
            "lang": language,
        }

        try:
            response = httpx.get(url, params=params, timeout=10)
            data = response.json()
            if data.get("weather"):
                for i in data["weather"]:
                    i.pop("hourly")

            return f"The weather for {city} for today and the next 2 days: {json.dumps(data, ensure_ascii=False)}"
        except Exception as e:
            return f"Failed to get the weather for {city} for today and the next 2 days: {str(e)}"



import asyncio
import platform
import uuid
from os import getenv
from os.path import basename, pathsep
from typing import Any, Callable, Optional, TypeVar

import typer
from distro import name as distro_name

from .const import DEFAULT_OS_NAME, DEFAULT_SHELL_NAME

T = TypeVar("T", int, float, str, bool)


def option_callback(func: Callable) -> Callable:  # pragma: no cover
    """
    A decorator for Typer option callbacks that ensures the application exits
    after the callback function is executed.

    Args:
        func (Callable): The callback classmethod to wrap.
    """

    def wrapper(cls, value: T) -> T:
        if not value:
            return value
        func(cls, value)
        raise typer.Exit()

    return wrapper


def detect_os(config: dict[str, Any]) -> str:
    """Detect operating system + version based on config or system info."""
    os_name_config = config.get("OS_NAME", DEFAULT_OS_NAME)
    if os_name_config != DEFAULT_OS_NAME:
        return os_name_config

    current_platform = platform.system()
    if current_platform == "Linux":
        return "Linux/" + distro_name(pretty=True)
    if current_platform == "Windows":
        return "Windows " + platform.release()
    if current_platform == "Darwin":
        return "Darwin/MacOS " + platform.mac_ver()[0]
    return current_platform


def detect_shell(config: dict[str, Any]) -> str:
    """Detect shell name based on config or environment."""
    shell_name_config = config.get("SHELL_NAME", DEFAULT_SHELL_NAME)
    if shell_name_config != DEFAULT_SHELL_NAME:
        return shell_name_config

    current_platform = platform.system()
    if current_platform in ("Windows", "nt"):
        # Basic check for PowerShell based on environment variables
        is_powershell = len(getenv("PSModulePath", "").split(pathsep)) >= 3
        return "powershell.exe" if is_powershell else "cmd.exe"

    # For Linux/MacOS, check SHELL environment variable
    return basename(getenv("SHELL") or "/bin/sh")


def filter_command(command: str) -> Optional[str]:
    """Filter out unwanted characters from command

    The LLM may return commands in markdown format with code blocks.
    This method removes markdown formatting from the command.
    It handles various formats including:
    - Commands surrounded by ``` (plain code blocks)
    - Commands with language specifiers like ```bash, ```zsh, etc.
    - Commands with specific examples like ```ls -al```

    example:
    ```bash\nls -la\n``` ==> ls -al
    ```zsh\nls -la\n``` ==> ls -al
    ```ls -la``` ==> ls -la
    ls -la ==> ls -la
    ```\ncd /tmp\nls -la\n``` ==> cd /tmp\nls -la
    ```bash\ncd /tmp\nls -la\n``` ==> cd /tmp\nls -la
    ```plaintext\nls -la\n``` ==> ls -la
    """
    if not command or not command.strip():
        return ""

    # Handle commands that are already without code blocks
    if "```" not in command:
        return command.strip()

    # Handle code blocks with or without language specifiers
    lines = command.strip().split("\n")

    # Check if it's a single-line code block like ```ls -al```
    if len(lines) == 1 and lines[0].startswith("```") and lines[0].endswith("```"):
        return lines[0][3:-3].strip()

    # Handle multi-line code blocks
    if lines[0].startswith("```"):
        # Remove the opening ``` line (with or without language specifier)
        content_lines = lines[1:]

        # If the last line is a closing ```, remove it
        if content_lines and content_lines[-1].strip() == "```":
            content_lines = content_lines[:-1]

        # Join the remaining lines and strip any extra whitespace
        return "\n".join(line.strip() for line in content_lines if line.strip())
    else:
        # If the first line doesn't start with ```, return the entire command without the ``` characters
        return command.strip().replace("```", "")


def str2bool(value: Any) -> bool:
    """Convert a string representation of truth to true (1) or false (0).
    True values are 'y', 'yes', 't', 'true', 'on', and '1';
    false values are 'n', 'no', 'f', 'false', 'off', and '0'.
    Raises ValueError if 'value' is anything else.
    """
    if value in {False, True}:
        return bool(value)

    if not isinstance(value, str):
        return value

    norm = value.strip().lower()

    if norm in {"1", "true", "t", "yes", "y", "on"}:
        return True

    if norm in {"0", "false", "f", "no", "n", "off"}:
        return False

    # Handle empty strings and other invalid values
    raise ValueError(f"Invalid boolean value: {value}")


def get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    """
    Get the current event loop or create a new one if it doesn't exist.
    Compatible with Python 3.10+.

    Returns:
        asyncio.AbstractEventLoop: The current event loop or a new one if it doesn't exist.
    """
    try:
        # Try to get the current running event loop
        return asyncio.get_running_loop()
    except RuntimeError:
        # No running event loop, try to get the current loop first
        # (for compatibility with environments that pre-set one).
        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            # No current loop available (e.g. Python 3.14+ behavior),
            # create and register a new loop for this thread.
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop


def gen_tool_call_id() -> str:
    """Generate a unique tool call id"""
    return f"yaicli_{uuid.uuid4()}"



import json
from typing import Generator, Optional, Union, overload

from openai._streaming import Stream
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice as ChoiceChunk

from ...schemas import LLMResponse, ToolCall
from .openai_provider import OpenAIProvider


class ChatglmProvider(OpenAIProvider):
    """Chatglm provider support"""

    DEFAULT_BASE_URL = "https://open.bigmodel.cn/api/paas/v4/"

    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_tokens": "MAX_TOKENS",
        "do_sample": "DO_SAMPLE",
        "extra_body": "EXTRA_BODY",
        "frequency_penalty": "FREQUENCY_PENALTY",
    }

    def _handle_normal_response(self, response: ChatCompletion) -> Generator[LLMResponse, None, None]:
        """Handle normal (non-streaming) response
        Support both openai capabilities and chatglm

        Returns:
            LLMContent object with:
            - reasoning: The thinking/reasoning content (if any)
            - content: The normal response content
        """
        choice = response.choices[0]
        content = choice.message.content or ""  # type: ignore
        reasoning = choice.message.reasoning_content  # type: ignore
        finish_reason = choice.finish_reason
        tool_call: Optional[ToolCall] = None

        # Check if the response contains reasoning content
        if "<think>" in content and "</think>" in content:
            # Extract reasoning content
            content = content.lstrip()
            if content.startswith("<think>"):
                think_end = content.find("</think>")
                if think_end != -1:
                    reasoning = content[7:think_end].strip()  # Start after <think>
                    # Remove the <think> block from the main content
                    content = content[think_end + 8 :].strip()  # Start after </think>
        # Check if the response contains reasoning content in model_extra
        elif hasattr(choice.message, "model_extra") and choice.message.model_extra:  # type: ignore
            model_extra = choice.message.model_extra  # type: ignore
            reasoning = self._get_reasoning_content(model_extra)
        if finish_reason == "tool_calls":
            if '{"index":' in content or '"tool_calls":' in content:
                # Tool call data may in content after the <think> block
                # >/n{"index": 0, "tool_call_id": "call_1", "function": {"name": "name", "arguments": "{}"}, "output": null}
                tool_index = content.find('{"index":')
                if tool_index != -1:
                    tmp_content = content[tool_index:]
                    # Tool call data may in content after the <think> block
                    try:
                        choice = self.parse_choice_from_content(tmp_content, Choice)
                    except ValueError:
                        pass
            if hasattr(choice, "message") and hasattr(choice.message, "tool_calls") and choice.message.tool_calls:  # type: ignore
                tool = choice.message.tool_calls[0]  # type: ignore
                tool_call = ToolCall(tool.id, tool.function.name or "", tool.function.arguments)

        yield LLMResponse(reasoning=reasoning, content=content, finish_reason=finish_reason, tool_call=tool_call)

    def _handle_stream_response(self, response: Stream[ChatCompletionChunk]) -> Generator[LLMResponse, None, None]:
        """Handle streaming response
        Support both openai capabilities and chatglm

        Returns:
            Generator yielding LLMContent objects with:
            - reasoning: The thinking/reasoning content (if any)
            - content: The normal response content
        """
        full_reasoning = ""
        full_content = ""
        content = ""
        reasoning = ""
        tool_id = ""
        tool_call_name = ""
        arguments = ""
        tool_call: Optional[ToolCall] = None
        for chunk in response:
            # Check if the response contains reasoning content
            choice = chunk.choices[0]  # type: ignore
            delta = choice.delta
            finish_reason = choice.finish_reason

            # Concat content
            content = delta.content or ""
            full_content += content

            # Concat reasoning
            reasoning = self._get_reasoning_content(delta)
            full_reasoning += reasoning or ""

            if finish_reason == "tool_calls" or ('{"index":' in content or '"tool_calls":' in content):
                # Tool call data may in content after the <think> block
                # >/n{"index": 0, "tool_call_id": "call_1", "function": {"name": "name", "arguments": "{}"}, "output": null}
                tool_index = full_content.find('{"index":')
                if tool_index != -1:
                    tmp_content = full_content[tool_index:]
                    try:
                        choice = self.parse_choice_from_content(tmp_content, ChoiceChunk)
                    except ValueError:
                        pass
            if hasattr(choice.delta, "tool_calls") and choice.delta.tool_calls:  # type: ignore
                # Handle tool calls
                tool_id = choice.delta.tool_calls[0].id or ""  # type: ignore
                for tool in choice.delta.tool_calls:  # type: ignore
                    if not tool.function:
                        continue
                    tool_call_name = tool.function.name or ""
                    arguments += tool.function.arguments or ""
                tool_call = ToolCall(tool_id, tool_call_name, arguments)
            yield LLMResponse(reasoning=reasoning, content=content, tool_call=tool_call, finish_reason=finish_reason)

    @overload
    def parse_choice_from_content(self, content: str, choice_class: type[ChoiceChunk] = ChoiceChunk) -> "ChoiceChunk":
        """
        Parse the choice from the content after <think>...</think> block.
        Args:
            content: The content from the LLM response
        Returns:
            The choice object
        Raises ValueError if the content is not valid JSON
        """

    @overload
    def parse_choice_from_content(self, content: str, choice_class: type[Choice] = Choice) -> "Choice":
        """
        Parse the choice from the content after <think>...</think> block.
        Args:
            content: The content from the LLM response
        Returns:
            The choice object
        Raises ValueError if the content is not valid JSON
        """

    def parse_choice_from_content(
        self, content: str, choice_class: type[Union[Choice, ChoiceChunk]] = Choice
    ) -> Union[Choice, ChoiceChunk]:
        """
        Parse the choice from the content after <think>...</think> block.
        Args:
            content: The content from the LLM response
        Returns:
            The choice object
        Raises ValueError if the content is not valid JSON
        """
        try:
            content_dict = json.loads(content)
        except json.JSONDecodeError:
            raise ValueError(f"Invalid message from LLM: {content}")
        try:
            return choice_class.model_validate(content_dict)
        except Exception as e:
            raise ValueError(f"Invalid message from LLM: {content}") from e



import importlib
import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, Generator, List

from ..schemas import ChatMessage, LLMResponse
from ..utils import option_callback

warnings.filterwarnings("ignore", category=DeprecationWarning)


class Provider(ABC):
    """Base abstract class for LLM providers"""

    APP_NAME = "yaicli"
    APP_REFERER = "https://github.com/belingud/yaicli"

    @abstractmethod
    def completion(
        self,
        messages: List[ChatMessage],
        stream: bool = False,
    ) -> Generator[LLMResponse, None, None]:
        """
        Send a completion request to the LLM provider

        Args:
            messages: List of message objects representing the conversation
            stream: Whether to stream the response

        Returns:
            Generator yielding LLMResponse objects
        """
        pass

    @abstractmethod
    def detect_tool_role(self) -> str:
        """Return the role that should be used for tool responses"""
        pass

    def _convert_messages(self, messages: List[ChatMessage]) -> List[Dict[str, Any]]:
        """Convert a list of ChatMessage objects to a list of OpenAI message format.

        Args:
            messages: List of ChatMessage

        Returns:
            List of OpenAI message format
        """
        converted_messages = []
        for msg in messages:
            message: Dict[str, Any] = {"role": msg.role, "content": msg.content or ""}

            if msg.name:
                message["name"] = msg.name

            if msg.role == "assistant" and msg.tool_calls:
                message["tool_calls"] = [
                    {"id": tc.id, "type": "function", "function": {"name": tc.name, "arguments": tc.arguments}}
                    for tc in msg.tool_calls
                ]

            if msg.role == "tool" and msg.tool_call_id:
                message["tool_call_id"] = msg.tool_call_id

            # Handle reasoning content for assistant messages
            if msg.role == "assistant" and msg.reasoning:
                message["reasoning"] = msg.reasoning

            converted_messages.append(message)

        return converted_messages

    @staticmethod
    def filter_excluded_params(
        params: Dict[str, Any],
        config: dict,
        verbose: bool = False,
        console=None,
    ) -> Dict[str, Any]:
        """Filter out excluded parameters from completion params.

        Args:
            params: The completion parameters dict to filter
            config: The config dict containing EXCLUDE_PARAMS
            verbose: Whether to log excluded parameters
            console: Console instance for logging (optional)

        Returns:
            Filtered params dict with excluded parameters removed

        Example:
            >>> params = {"temperature": 0.5, "top_p": 1.0, "model": "gpt-4"}
            >>> config = {"EXCLUDE_PARAMS": "temperature,top_p"}
            >>> Provider.filter_excluded_params(params, config)
            {"model": "gpt-4"}
        """
        exclude_str = config.get("EXCLUDE_PARAMS", "")
        if not exclude_str or not exclude_str.strip():
            return params

        # Parse exclude list: strip whitespace, normalize to lowercase
        exclude_list = [p.strip().lower() for p in exclude_str.split(",") if p.strip()]

        if not exclude_list:
            return params

        # Track what was excluded for logging
        excluded_keys = [k for k in params.keys() if k.lower() in exclude_list]

        # Filter params using case-insensitive comparison
        filtered = {k: v for k, v in params.items() if k.lower() not in exclude_list}

        # Log if verbose and something was excluded
        if verbose and excluded_keys and console:
            console.print(
                f"[yellow]Excluded parameters:[/yellow] {', '.join(excluded_keys)}",
                style="dim",
            )

        return filtered


class ProviderFactory:
    """Factory to create LLM provider instances"""

    providers_map = {
        "ai21": (".providers.ai21_provider", "AI21Provider"),
        "anthropic": (".providers.anthropic_provider", "AnthropicProvider"),
        "anthropic-bedrock": (".providers.anthropic_provider", "AnthropicBedrockProvider"),
        "anthropic-vertex": (".providers.anthropic_provider", "AnthropicVertexProvider"),
        "bailian": (".providers.bailian_provider", "BailianProvider"),
        "bailian-intl": (".providers.bailian_provider", "BailianIntlProvider"),
        "cerebras": (".providers.cerebras_provider", "CerebrasProvider"),
        "chatglm": (".providers.chatglm_provider", "ChatglmProvider"),
        "chutes": (".providers.chutes_provider", "ChutesProvider"),
        "cohere": (".providers.cohere_provider", "CohereProvider"),
        "cohere-bedrock": (".providers.cohere_provider", "CohereBadrockProvider"),
        "cohere-sagemaker": (".providers.cohere_provider", "CohereSagemakerProvider"),
        "deepseek": (".providers.deepseek_provider", "DeepSeekProvider"),
        "doubao": (".providers.doubao_provider", "DoubaoProvider"),
        "fireworks": (".providers.fireworks_provider", "FireworksProvider"),
        "gemini": (".providers.gemini_provider", "GeminiProvider"),
        "groq": (".providers.groq_provider", "GroqProvider"),
        "huggingface": (".providers.huggingface_provider", "HuggingFaceProvider"),
        "infini-ai": (".providers.infiniai_provider", "InfiniAIProvider"),
        "longcat": (".providers.longcat_provider", "LongCatProvider"),
        "longcat-anthropic": (".providers.longcat_provider", "LongCatAnthropicProvider"),
        "minimax": (".providers.minimax_provider", "MinimaxProvider"),
        "mistral": (".providers.mistral_provider", "MistralProvider"),
        "modelscope": (".providers.modelscope_provider", "ModelScopeProvider"),
        "moonshot": (".providers.moonshot_provider", "MoonshotProvider"),
        "nvida": (".providers.nvida_provider", "NvidiaProvider"),
        "ollama": (".providers.ollama_provider", "OllamaProvider"),
        "openai": (".providers.openai_provider", "OpenAIProvider"),
        "openai-azure": (".providers.openai_provider", "OpenAIAzure"),
        "openai-compatible": (".providers.openai_compatible_provider", "OpenAICompatibleProvider"),
        "openrouter": (".providers.openrouter_provider", "OpenRouterProvider"),
        "sambanova": (".providers.sambanova_provider", "SambanovaProvider"),
        "siliconflow": (".providers.siliconflow_provider", "SiliconFlowProvider"),
        "spark": (".providers.spark_provider", "SparkProvider"),
        "targon": (".providers.targon_provider", "TargonProvider"),
        "together": (".providers.together_provider", "TogetherProvider"),
        "vertexai": (".providers.vertexai_provider", "VertexAIProvider"),
        "xai": (".providers.xai_provider", "XaiProvider"),
        "yi": (".providers.yi_provider", "YiProvider"),
    }

    @classmethod
    def create_provider(cls, provider_type: str, verbose: bool = False, **kwargs) -> Provider:
        """Create a provider instance based on provider type

        Args:
            provider_type: The type of provider to create
            **kwargs: Additional parameters to pass to the provider

        Returns:
            A Provider instance
        """
        provider_type = provider_type.lower()
        if provider_type not in cls.providers_map:
            raise ValueError(f"Unknown provider: {provider_type}")

        module_path, class_name = cls.providers_map[provider_type]
        module = importlib.import_module(module_path, package="yaicli.llms")
        return getattr(module, class_name)(verbose=verbose, **kwargs)

    @classmethod
    @option_callback
    def list_providers(cls, _: Any) -> None:
        """List the available providers and exit."""
        print("Available PROVIDERS:")
        print("--------------------")
        for i in cls.providers_map:
            print(i)
        print("--------------------")



import fnmatch
import json
import os
from pathlib import Path
from typing import List

from yaicli.function_schema import OpenAISchema
from pydantic import Field


class Function(OpenAISchema):
    """
    Search for files matching a pattern and return results as JSON.
    """

    search_path: str = Field(
        ...,
        json_schema_extra={
            "example": "/path/to/search",
        },
        description="Directory path to search in.",
    )
    pattern: str = Field(
        ...,
        json_schema_extra={
            "example": "*.py",
        },
        description="File name pattern to match (supports wildcards like *.txt, test_*.py).",
    )
    exclude_patterns: List[str] = Field(
        default_factory=list,
        json_schema_extra={
            "example": ["*.pyc", "__pycache__", ".git"],
        },
        description="Patterns to exclude from search results.",
    )
    max_results: int = Field(
        default=1000,
        json_schema_extra={
            "example": 1000,
        },
        description="Maximum number of results to return (default: 1000).",
    )

    class Config:
        title = "fs_search_files"

    @classmethod
    def execute(
        cls, search_path: str, pattern: str, exclude_patterns: List[str] | None = None, max_results: int = 1000
    ) -> str:
        """
        Search for files matching a pattern and return results as JSON.

        Args:
            search_path: Directory to search in.
            pattern: File name pattern (e.g., "*.py", "test_*.txt").
            exclude_patterns: Patterns to exclude.
            max_results: Maximum number of results.

        Returns:
            str: JSON string with search results.
        """
        if exclude_patterns is None:
            exclude_patterns = []

        result = {
            "search_path": search_path,
            "pattern": pattern,
            "success": False,
            "matches": [],
            "total_scanned": 0,
            "excluded_count": 0,
            "match_count": 0,
            "truncated": False,
            "error": None,
        }

        try:
            path = Path(search_path).expanduser().resolve()

            # Check if path exists
            if not path.exists():
                result["error"] = "Search path does not exist"
                return json.dumps(result, ensure_ascii=False, indent=2)

            if not path.is_dir():
                result["error"] = "Search path is not a directory"
                return json.dumps(result, ensure_ascii=False, indent=2)

            matches = []
            total_scanned = 0
            excluded_count = 0

            # Walk through directory tree
            for root, dirs, files in os.walk(path):
                root_path = Path(root)

                # Filter out excluded directories
                dirs_to_remove = []
                for d in dirs:
                    if cls._should_exclude(d, exclude_patterns):
                        dirs_to_remove.append(d)
                        excluded_count += 1

                for d in dirs_to_remove:
                    dirs.remove(d)

                # Check files
                for file_name in files:
                    total_scanned += 1

                    # Check if file should be excluded
                    if cls._should_exclude(file_name, exclude_patterns):
                        excluded_count += 1
                        continue

                    # Check if file matches pattern
                    if fnmatch.fnmatch(file_name, pattern):
                        file_path = root_path / file_name
                        rel_path = file_path.relative_to(path)

                        match_info = {
                            "name": file_name,
                            "path": str(rel_path),
                            "full_path": str(file_path),
                            "directory": str(rel_path.parent) if rel_path.parent != Path(".") else ".",
                            "size": None,
                        }

                        try:
                            match_info["size"] = file_path.stat().st_size
                        except Exception:
                            pass

                        matches.append(match_info)

                        # Check max results limit
                        if len(matches) >= max_results:
                            result["truncated"] = True
                            break

                if len(matches) >= max_results:
                    break

            result["success"] = True
            result["matches"] = matches
            result["total_scanned"] = total_scanned
            result["excluded_count"] = excluded_count
            result["match_count"] = len(matches)

            return json.dumps(result, ensure_ascii=False, indent=2)

        except PermissionError:
            result["error"] = "Permission denied"
            return json.dumps(result, ensure_ascii=False, indent=2)

        except Exception as e:
            result["error"] = str(e)
            return json.dumps(result, ensure_ascii=False, indent=2)

    @staticmethod
    def _should_exclude(name: str, exclude_patterns: List[str]) -> bool:
        """Check if a name matches any exclude pattern."""
        for pattern in exclude_patterns:
            if fnmatch.fnmatch(name, pattern):
                return True
        return False



import json
import os
import platform
import shutil
from pathlib import Path

from yaicli.function_schema import OpenAISchema
from pydantic import Field


class Function(OpenAISchema):
    """
    Get information about the filesystem and system configuration as JSON.
    """

    include_disk_usage: bool = Field(
        default=True,
        json_schema_extra={
            "example": True,
        },
        description="Include disk usage information for common mount points.",
    )

    class Config:
        title = "fs_get_filesystem_info"

    @classmethod
    def execute(cls, include_disk_usage: bool = True) -> str:
        """
        Get filesystem and system information as JSON.

        Args:
            include_disk_usage: Include disk usage information.

        Returns:
            str: JSON string with system and filesystem information.
        """
        result = {
            "success": True,
            "system": {},
            "directories": {},
            "environment": {},
            "disk_usage": [],
            "filesystem_limits": {},
            "error": None,
        }

        try:
            # System Information
            result["system"] = {
                "os": platform.system(),
                "os_version": platform.release(),
                "platform": platform.platform(),
                "architecture": platform.machine(),
                "processor": platform.processor(),
                "python_version": platform.python_version(),
            }

            # Current Directory Information
            result["directories"] = {"working_directory": str(Path.cwd()), "home_directory": str(Path.home())}

            # Environment Variables (selected)
            important_vars = ["PATH", "HOME", "USER", "SHELL", "TMPDIR", "TEMP", "TMP"]
            env_vars = {}
            for var in important_vars:
                value = os.environ.get(var)
                if value:
                    # Truncate long values
                    if len(value) > 100:
                        value = value[:100] + "..."
                    env_vars[var] = value
            result["environment"] = env_vars

            # Disk Usage
            if include_disk_usage:
                paths_to_check = [
                    Path.home(),
                    Path.cwd(),
                    Path("/tmp") if os.path.exists("/tmp") else None,
                    Path("/"),
                ]

                checked_devices = set()
                disk_usage_list = []

                for path in paths_to_check:
                    if path is None:
                        continue

                    try:
                        usage = shutil.disk_usage(path)

                        # Get device/mount point (avoid duplicates)
                        mount_point = cls._get_mount_point(path)
                        if mount_point in checked_devices:
                            continue
                        checked_devices.add(mount_point)

                        total = usage.total
                        used = usage.used
                        free = usage.free
                        percent = (used / total * 100) if total > 0 else 0

                        disk_usage_list.append(
                            {
                                "mount_point": mount_point,
                                "total": total,
                                "used": used,
                                "free": free,
                                "percent_used": round(percent, 2),
                            }
                        )

                    except Exception:
                        pass  # Skip paths that can't be accessed

                result["disk_usage"] = disk_usage_list

            # File System Limits
            try:
                if hasattr(os, "pathconf"):
                    name_max = os.pathconf("/", "PC_NAME_MAX")
                    path_max = os.pathconf("/", "PC_PATH_MAX")
                    result["filesystem_limits"] = {"max_filename_length": name_max, "max_path_length": path_max}
                else:
                    result["filesystem_limits"] = {"available": False, "message": "Not available on this platform"}
            except Exception:
                result["filesystem_limits"] = {"available": False, "message": "Unable to determine"}

            return json.dumps(result, ensure_ascii=False, indent=2)

        except Exception as e:
            result["success"] = False
            result["error"] = str(e)
            return json.dumps(result, ensure_ascii=False, indent=2)

    @staticmethod
    def _get_mount_point(path: Path) -> str:
        """Get the mount point for a given path."""
        path = path.resolve()
        while not path.is_mount():
            parent = path.parent
            if parent == path:
                break
            path = parent
        return str(path)



import sys
from typing import Annotated, Any, Optional

import typer

from .chat import FileChatManager
from .config import cfg
from .const import DEFAULT_CONFIG_INI, DefaultRoleNames, JustifyEnum
from .exceptions import YaicliError
from .functions import install_functions, print_functions, print_mcp, reinstall_functions
from .llms.provider import ProviderFactory
from .role import RoleManager

app = typer.Typer(
    name="yaicli",
    help="YAICLI - Your AI assistant in command line.",
    context_settings={"help_option_names": ["-h", "--help"]},
    pretty_exceptions_enable=False,  # Let the CLI handle errors gracefully
    rich_markup_mode="rich",  # Render rich text in help messages
)


def override_config(
    ctx: typer.Context,  # noqa: F841
    param: typer.CallbackParam,
    value: Any,
):
    """Override config with input value if value not equal to option default."""
    if value != param.default and isinstance(param.name, str):
        cfg[param.name.upper()] = value
    return value


class LLMOptions:
    model = typer.Option(
        "",
        "--model",
        "-M",
        help="Specify the model to use.",
        rich_help_panel="LLM Options",
        callback=override_config,
    )

    temperature = typer.Option(
        cfg["TEMPERATURE"],
        "--temperature",
        "-T",
        help="Specify the temperature to use.",
        rich_help_panel="LLM Options",
        min=0.0,
        max=2.0,
        callback=override_config,
    )

    top_p = typer.Option(
        cfg["TOP_P"],
        "--top-p",
        "-P",
        help="Specify the top-p to use.",
        rich_help_panel="LLM Options",
        min=0.0,
        max=1.0,
        callback=override_config,
    )

    max_tokens = typer.Option(
        cfg["MAX_TOKENS"],
        "--max-tokens",
        help="Specify the max tokens to use.",
        rich_help_panel="LLM Options",
        min=1,
        callback=override_config,
    )

    stream = typer.Option(
        cfg["STREAM"],
        "--stream/--no-stream",
        help=f"Specify whether to stream the response. [dim](default: {'stream' if cfg['STREAM'] else 'no-stream'})[/dim]",
        rich_help_panel="LLM Options",
        show_default=False,
        callback=override_config,
    )


class RoleOptions:
    role = typer.Option(
        cfg["DEFAULT_ROLE"],
        "--role",
        "-r",
        help="Specify the assistant role to use.",
        rich_help_panel="Role Options",
        callback=RoleManager.check_id_ok,
    )

    create_role = typer.Option(
        "",
        "--create-role",
        help="Create a new role with the specified name.",
        rich_help_panel="Role Options",
        callback=RoleManager.create_role_option,
    )

    delete_role = typer.Option(
        "",
        "--delete-role",
        help="Delete a role with the specified name.",
        rich_help_panel="Role Options",
        callback=RoleManager.delete_role_option,
    )

    list_roles = typer.Option(
        False,
        "--list-roles",
        help="List all available roles.",
        rich_help_panel="Role Options",
        callback=RoleManager.print_list_option,
    )

    show_role = typer.Option(
        "",
        "--show-role",
        help="Show the role with the specified name.",
        rich_help_panel="Role Options",
        callback=RoleManager.show_role_option,
    )


class ChatOptions:
    chat = typer.Option(
        False,
        "--chat",
        "-c",
        help="Start in interactive chat mode.",
        rich_help_panel="Chat Options",
    )

    list_chats = typer.Option(
        False,
        "--list-chats",
        help="List saved chat sessions.",
        rich_help_panel="Chat Options",
        callback=FileChatManager.print_list_option,
    )


class ShellOptions:
    shell = typer.Option(
        False,
        "--shell",
        "-s",
        help="Generate and optionally execute a shell command (non-interactive).",
        rich_help_panel="Shell Options",
    )


class CodeOptions:
    code = typer.Option(
        False,
        "--code",
        help="Generate code in plaintext (non-interactive).",
        rich_help_panel="Code Options",
    )


class OtherOptions:
    verbose = typer.Option(
        False,
        "--verbose",
        "-V",
        help="Show verbose output (e.g., loaded config).",
        rich_help_panel="Other Options",
    )

    template = typer.Option(
        False,
        "--template",
        help="Show the default config file template and exit.",
        rich_help_panel="Other Options",
    )

    list_providers = typer.Option(
        False,
        "--list-providers",
        help="List the available providers and exit.",
        rich_help_panel="Other Options",
        callback=ProviderFactory.list_providers,
    )

    show_reasoning = typer.Option(
        cfg["SHOW_REASONING"],
        "--show-reasoning/--hide-reasoning",
        help=f"Show reasoning content from the LLM. [dim](default: {'show' if cfg['SHOW_REASONING'] else 'hide'})[/dim]",
        rich_help_panel="Other Options",
        show_default=False,
        callback=override_config,
    )

    justify = typer.Option(
        cfg["JUSTIFY"],
        "--justify",
        "-j",
        help="Specify the justify to use.",
        rich_help_panel="Other Options",
        callback=override_config,
    )


class MCPOptions:
    enable_mcp = typer.Option(
        cfg["ENABLE_MCP"],
        "--enable-mcp/--disable-mcp",
        help=f"Enable/disable MCP in API requests [dim](default: {'enabled' if cfg['ENABLE_MCP'] else 'disabled'})[/dim]",
        rich_help_panel="MCP Options",
        callback=override_config,
    )

    show_mcp_output = typer.Option(
        cfg["SHOW_MCP_OUTPUT"],
        "--show-mcp-output/--hide-mcp-output",
        help=f"Show the output of MCP [dim](default: {'show' if cfg['SHOW_MCP_OUTPUT'] else 'hide'})[/dim]",
        rich_help_panel="MCP Options",
        show_default=False,
        callback=override_config,
    )
    list_mcp = typer.Option(
        False,
        "--list-mcp",
        help="List all available mcp.",
        rich_help_panel="MCP Options",
        callback=print_mcp,
    )


class FunctionOptions:
    install_functions = typer.Option(
        False,
        "--install-functions",
        help="Install default functions.",
        rich_help_panel="Function Options",
        callback=install_functions,
    )

    list_functions = typer.Option(
        False,
        "--list-functions",
        help="List all available functions.",
        rich_help_panel="Function Options",
        callback=print_functions,
    )

    reinstall_functions = typer.Option(
        False,
        "--reinstall-functions",
        help="Reinstall builtin functions (overwrites existing builtin files, preserves custom ones).",
        rich_help_panel="Function Options",
        callback=reinstall_functions,
    )

    enable_functions = typer.Option(
        cfg["ENABLE_FUNCTIONS"],
        "--enable-functions/--disable-functions",
        help=f"Enable/disable function calling in API requests [dim](default: {'enabled' if cfg['ENABLE_FUNCTIONS'] else 'disabled'})[/dim]",
        rich_help_panel="Function Options",
        show_default=False,
        callback=override_config,
    )

    show_function_output = typer.Option(
        cfg["SHOW_FUNCTION_OUTPUT"],
        "--show-function-output/--hide-function-output",
        help=f"Show the output of functions [dim](default: {'show' if cfg['SHOW_FUNCTION_OUTPUT'] else 'hide'})[/dim]",
        rich_help_panel="Function Options",
        show_default=False,
        callback=override_config,
    )


@app.command()
def main(
    ctx: typer.Context,
    prompt: Annotated[
        Optional[str], typer.Argument(help="The prompt to send to the LLM. Reads from stdin if available.")
    ] = None,
    # ------------------- LLM Options -------------------
    model: str = LLMOptions.model,  # noqa: F841
    temperature: float = LLMOptions.temperature,  # noqa: F841
    top_p: float = LLMOptions.top_p,  # noqa: F841
    max_tokens: int = LLMOptions.max_tokens,  # noqa: F841
    stream: bool = LLMOptions.stream,  # noqa: F841
    # ------------------- Role Options -------------------
    role: str = RoleOptions.role,
    create_role: str = RoleOptions.create_role,  # noqa: F841
    delete_role: str = RoleOptions.delete_role,  # noqa: F841
    list_roles: bool = RoleOptions.list_roles,  # noqa: F841
    show_role: str = RoleOptions.show_role,  # noqa: F841
    # ------------------- Chat Options -------------------
    chat: bool = ChatOptions.chat,
    list_chats: bool = ChatOptions.list_chats,  # noqa: F841
    # ------------------- Shell Options -------------------
    shell: bool = ShellOptions.shell,
    # ------------------- Code Options -------------------
    code: bool = CodeOptions.code,
    # ------------------- Other Options -------------------
    verbose: bool = OtherOptions.verbose,
    template: bool = OtherOptions.template,
    list_providers: bool = OtherOptions.list_providers,  # noqa: F841
    show_reasoning: bool = OtherOptions.show_reasoning,  # noqa: F841
    justify: JustifyEnum = OtherOptions.justify,  # noqa: F841
    # ------------------- Function Options -------------------
    install_functions: bool = FunctionOptions.install_functions,  # noqa: F841
    reinstall_functions: bool = FunctionOptions.reinstall_functions,  # noqa: F841
    list_functions: bool = FunctionOptions.list_functions,  # noqa: F841
    enable_functions: bool = FunctionOptions.enable_functions,  # noqa: F841
    show_function_output: bool = FunctionOptions.show_function_output,  # noqa: F841
    # ------------------- MCP Options -------------------
    enable_mcp: bool = MCPOptions.enable_mcp,  # noqa: F841
    show_mcp_output: bool = MCPOptions.show_mcp_output,  # noqa: F841
    list_mcp: bool = MCPOptions.list_mcp,  # noqa: F841
):
    """YAICLI: Your AI assistant in the command line.

    Call with a PROMPT to get a direct answer, use --shell to execute as command, or use --chat for an interactive session.

    Example:
        ai "What is the capital of France?"
        ai --code "Write a fibonacci generator in Python"
        ai --chat "What's the meaning of life"  # Start a chat session with a title
        ai --chat  # Start a temporary chat session
        ai --shell "Start docker image python:3.13.5-alpine3.22 with ports 8080 and map current dir to /app in container"
    """
    if template:
        print(DEFAULT_CONFIG_INI)
        raise typer.Exit()

    # # Combine prompt argument with stdin content if available
    final_prompt = prompt
    if not sys.stdin.isatty():
        stdin_content = sys.stdin.read().strip()
        if stdin_content:
            if final_prompt:
                # Prepend stdin content to the argument prompt
                final_prompt = f"{stdin_content}\n\n{final_prompt}"
            else:
                final_prompt = stdin_content
        # prompt_toolkit will raise EOFError if stdin is redirected
        # Set chat to False to prevent starting interactive mode.
        if chat:
            print("Warning: --chat is ignored when stdin was redirected.")
            chat = False
    if not any([final_prompt, chat]):
        print(ctx.get_help())
        return

    # # Use build-in role for --shell or --code mode
    if role and role != DefaultRoleNames.DEFAULT and (shell or code):
        print("Warning: --role is ignored when --shell or --code is used.")
        role = DefaultRoleNames.DEFAULT

    from yaicli.cli import CLI

    role = CLI.evaluate_role_name(code, shell, role)

    try:
        # Instantiate the main CLI class with the specified role
        cli = CLI(verbose=verbose, role=role)
        # Run the appropriate mode
        cli.run(
            chat=chat,
            shell=shell,
            code=code,
            user_input=final_prompt,
        )
    except YaicliError as e:
        typer.echo(f"YAICLI Error: {e}")
    except (typer.Abort, typer.Exit):
        pass
    except Exception as e:
        typer.echo(f"Unknown error: {e}")


if __name__ == "__main__":
    app()



from typing import Generator, List, Union

from ..config import cfg
from ..console import get_console
from ..schemas import ChatMessage, LLMResponse, RefreshLive, ToolCall
from ..tools import execute_tool_call
from ..tools.mcp import MCP_TOOL_NAME_PREFIX
from .provider import ProviderFactory


class LLMClient:
    """
    LLM Client that coordinates provider interactions and tool calling

    This class handles the higher level logic of:
    1. Getting responses from LLM providers
    2. Managing tool calls and their execution
    3. Handling conversation flow with tools
    """

    __slots__ = ("config", "verbose", "console", "enable_function", "enable_mcp", "max_tool_call_depth", "provider")

    def __init__(
        self,
        provider_name: str,
        config: dict = cfg,
        verbose: bool = False,
        **kwargs,
    ):
        """
        Initialize LLM client

        Args:
            provider_name: Name of the provider to use, default to openai if not known
            config: Configuration dictionary
            verbose: Whether to enable verbose logging
        """
        self.config = config
        self.verbose = verbose
        self.console = get_console()
        self.enable_function = self.config["ENABLE_FUNCTIONS"]
        self.enable_mcp = self.config["ENABLE_MCP"]

        # Use provided provider or create one
        if provider_name not in ProviderFactory.providers_map:
            self.console.print(f"Provider {provider_name} not found, using openai as default", style="yellow")
            provider_name = "openai"
        self.provider = ProviderFactory.create_provider(provider_name, config=config, verbose=verbose, **kwargs)

        self.max_tool_call_depth = self.config["MAX_TOOL_CALL_DEPTH"]

    def completion_with_tools(
        self,
        messages: List[ChatMessage],
        stream: bool = False,
        recursion_depth: int = 0,
    ) -> Generator[Union[LLMResponse, RefreshLive], None, None]:
        """
        Get completion from provider with tool calling support

        Args:
            messages: List of messages for the conversation
            stream: Whether to stream the response
            recursion_depth: Current recursion depth for tool calls

        Yields:
            LLMResponse objects and control signals
        """
        if recursion_depth >= self.max_tool_call_depth:
            self.console.print(
                f"Maximum tool call depth ({self.max_tool_call_depth}) reached. Stopping further tool calls...",
                style="bold yellow",
            )
            return

        # Get completion from provider and collect response data
        assistant_response_content = ""
        assistant_reasoning_content = ""  # Collect reasoning content
        # Providers may return identical tool calls with the same ID in a single response during streaming
        tool_calls: dict[str, ToolCall] = {}

        # Stream responses and collect data
        for llm_response in self.provider.completion(messages, stream=stream):
            yield llm_response  # Forward response to caller

            # Collect content and tool calls for potential tool execution
            if llm_response.content:
                assistant_response_content += llm_response.content
            if llm_response.reasoning:  # Collect reasoning content
                assistant_reasoning_content += llm_response.reasoning
            if llm_response.tool_call and llm_response.tool_call.id not in tool_calls:
                tool_calls[llm_response.tool_call.id] = llm_response.tool_call

        # Always add assistant response to messages first
        assistant_message = ChatMessage(
            role="assistant",
            content=assistant_response_content,
            tool_calls=list(tool_calls.values()) if tool_calls else [],
            reasoning=assistant_reasoning_content if assistant_reasoning_content else None,  # Save reasoning
        )
        messages.append(assistant_message)

        # Check if we need to execute tools
        if not tool_calls or not (self.enable_function or self.enable_mcp):
            return

        # Filter valid tool calls based on enabled features
        valid_tool_calls = self._get_valid_tool_calls(tool_calls)
        if not valid_tool_calls:
            return

        # Execute tools and continue conversation
        yield from self._execute_tools_and_continue(messages, valid_tool_calls, stream, recursion_depth)

    def _get_valid_tool_calls(self, tool_calls: dict[str, ToolCall]) -> List[ToolCall]:
        """Filter tool calls based on enabled features"""
        valid_tool_calls = []

        for tool_call in tool_calls.values():
            if self.verbose:
                self.console.print(f"Raw tool call name: {tool_call.name}")
            is_mcp = tool_call.name.startswith(MCP_TOOL_NAME_PREFIX)

            if is_mcp and self.enable_mcp:
                valid_tool_calls.append(tool_call)
            elif not is_mcp and self.enable_function:
                valid_tool_calls.append(tool_call)

        return valid_tool_calls

    def _execute_tools_and_continue(
        self,
        messages: List[ChatMessage],
        tool_calls: List[ToolCall],
        stream: bool,
        recursion_depth: int,
    ) -> Generator[Union[LLMResponse, RefreshLive], None, None]:
        """Execute tool calls and continue the conversation"""
        # Signal that new content is coming
        yield RefreshLive()

        # Assistant message has already been added to messages in completion_with_tools

        # Execute each tool call and add results to messages
        tool_role = self.provider.detect_tool_role()

        for tool_call in tool_calls:
            function_result, _ = execute_tool_call(tool_call)

            messages.append(
                ChatMessage(
                    role=tool_role,
                    content=function_result,
                    name=tool_call.name,
                    tool_call_id=tool_call.id,
                )
            )

        # Continue the conversation with updated history
        yield from self.completion_with_tools(messages, stream=stream, recursion_depth=recursion_depth + 1)



from typing import Any, Dict

from huggingface_hub import InferenceClient

from .chatglm_provider import ChatglmProvider


class HuggingFaceProvider(ChatglmProvider):
    """
    HuggingFaceProvider is a provider for the HuggingFace API.
    """

    CLIENT_CLS = InferenceClient
    DEFAULT_PROVIDER = "auto"

    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_tokens": "MAX_TOKENS",
        "extra_body": "EXTRA_BODY",
        "frequency_penalty": "FREQUENCY_PENALTY",
    }

    def get_client_params(self) -> Dict[str, Any]:
        client_params = {
            "api_key": self.config["API_KEY"],
            "timeout": self.config["TIMEOUT"],
            "provider": self.config.get("HF_PROVIDER") or self.DEFAULT_PROVIDER,
        }
        if self.config["BASE_URL"]:
            client_params["base_url"] = self.config["BASE_URL"]
        if self.config["EXTRA_HEADERS"]:
            client_params["headers"] = {
                **self.config["EXTRA_HEADERS"],
                "X-Title": self.APP_NAME,
                "HTTP-Referer": self.APP_REFERER,
            }
        if self.config.get("BILL_TO"):
            client_params["bill_to"] = self.config["BILL_TO"]
        return client_params



import importlib.util
import json
import sys
from functools import wraps
from typing import TYPE_CHECKING, Callable, List, Optional

if TYPE_CHECKING:
    from yaicli.function_schema import OpenAISchema

from ..console import get_console
from ..const import FUNCTIONS_DIR

_deprecated_warned: set[str] = set()


def wrap_gemini_function(func: Callable) -> Callable:
    """Wrap a function to add a name and docstring"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        print(f"\033[94m@Function call: {wrapper.__name__}({json.dumps(kwargs) if kwargs else args})\033[0m")
        return func(*args, **kwargs)

    return wrapper


class Function:
    """Function description class"""

    def __init__(self, function: type["OpenAISchema"]):
        self.name = function.openai_schema["name"]
        self.description = function.openai_schema.get("description", "")
        self.parameters = function.openai_schema.get("parameters", {})
        self.execute = function.execute  # type: ignore
        self.func_cls = function


_func_name_map: Optional[dict[str, Function]] = None


def _check_deprecated_import(file_path) -> None:
    """Check if a function file uses the deprecated instructor import and print a warning."""
    name = file_path.name
    if name in _deprecated_warned:
        return
    try:
        source = file_path.read_text(encoding="utf-8")
    except Exception:
        return
    if "from instructor import" in source:
        _deprecated_warned.add(name)
        console = get_console()
        console.print(
            f"[yellow]WARNING:[/yellow] {name}: 'from instructor import OpenAISchema' is deprecated "
            f"and will be removed in a future version. "
            f"Please change to 'from yaicli.function_schema import OpenAISchema'. "
            f"Run [bold]ai --reinstall-functions[/bold] to update builtin functions."
        )


def get_func_name_map() -> dict[str, Function]:
    """Get function name map"""
    global _func_name_map
    if _func_name_map:
        return _func_name_map
    if not FUNCTIONS_DIR.exists():
        FUNCTIONS_DIR.mkdir(parents=True, exist_ok=True)
        return {}
    functions = []
    for file in FUNCTIONS_DIR.glob("*.py"):
        if file.name.startswith("_"):
            continue

        _check_deprecated_import(file)

        module_name = str(file).replace("/", ".").rstrip(".py")
        spec = importlib.util.spec_from_file_location(module_name, str(file))
        module = importlib.util.module_from_spec(spec)  # type: ignore
        sys.modules[module_name] = module
        spec.loader.exec_module(module)  # type: ignore

        if not hasattr(module.Function, "openai_schema") or not hasattr(module.Function, "anthropic_schema"):
            raise TypeError(
                f"Function {module_name} must be a subclass of yaicli.function_schema.OpenAISchema"
            )
        if not hasattr(module.Function, "execute"):
            raise TypeError(f"Function {module_name} must have an 'execute' classmethod")

        # Add to function list
        functions.append(Function(function=module.Function))

    # Cache the function list
    _func_name_map = {func.name: func for func in functions}
    return _func_name_map


def list_functions() -> list[Function]:
    """List all available buildin functions"""
    global _func_name_map
    if not _func_name_map:
        _func_name_map = get_func_name_map()

    return list(_func_name_map.values())


def get_function(name: str) -> Function:
    """Get a function by name

    Args:
        name: Function name

    Returns:
        Function execute method

    Raises:
        ValueError: If function not found
    """
    func_map = get_func_name_map()
    if name in func_map:
        return func_map[name]
    raise ValueError(f"Function {name!r} not found")


def get_functions_gemini_format() -> List[Callable]:
    """Get functions in gemini format"""
    gemini_functions = []
    for func_name, func in get_func_name_map().items():
        wrapped_func = wrap_gemini_function(func.execute)
        wrapped_func.__name__ = func_name
        wrapped_func.__doc__ = func.description
        gemini_functions.append(wrapped_func)
    return gemini_functions



import json
from typing import Any, Dict, Generator, List, Optional, Union, cast

from mistralai import (
    DocumentURLChunk,
    ImageURLChunk,
    Mistral,
    ReferenceChunk,
    TextChunk,
)
from mistralai.models import ChatCompletionResponse, CompletionEvent, ContentChunk
from mistralai.models import ToolCall as MistralToolCall
from mistralai.utils.eventstreaming import EventStream

from ...config import cfg
from ...console import get_console
from ...exceptions import MCPToolsError
from ...schemas import ChatMessage, LLMResponse, ToolCall
from ...tools import get_openai_mcp_tools, get_openai_schemas
from ...utils import gen_tool_call_id
from ..provider import Provider


class MistralProvider(Provider):
    """Mistral provider implementation based on mistralai library"""

    CLIENT_CLS = Mistral
    DEFAULT_BASE_URL = "https://api.mistral.ai/v1"

    def __init__(self, config: dict = cfg, verbose: bool = False, **kwargs):
        """Initialize Mistral provider

        Args:
            config: Configuration dictionary
            verbose: Whether to enable verbose logging
        """
        self.config = config
        self.verbose = verbose
        self.enable_functions = config["ENABLE_FUNCTIONS"]
        self.enable_mcp = config["ENABLE_MCP"]
        self.client = Mistral(**self.get_client_params())
        self.console = get_console()

    def get_client_params(self) -> Dict[str, Any]:
        """Get client parameters for Mistral

        Returns:
            Dict[str, Any]
        """
        client_params = {
            "api_key": self.config["API_KEY"],
            "timeout_ms": self.config["TIMEOUT"] * 1000,  # Mistral uses milliseconds
        }
        if self.config["BASE_URL"]:
            client_params["server_url"] = self.config["BASE_URL"]
        if self.config.get("SERVER"):
            client_params["server"] = self.config["SERVER"]
        return client_params

    def get_completion_params(self) -> Dict[str, Any]:
        """Get completion parameters for Mistral

        Returns:
            Dict[str, Any]
        """
        params = {
            "model": self.config["MODEL"],
            "temperature": self.config["TEMPERATURE"],
            "top_p": self.config["TOP_P"],
            "max_tokens": self.config["MAX_TOKENS"],
            "stream": self.config["STREAM"],
            "http_headers": {
                "X-Title": self.APP_NAME,
                "HTTP_Referer": self.APP_REFERER,
            },
            "frequency_penalty": self.config["FREQUENCY_PENALTY"],
        }
        if self.config["EXTRA_HEADERS"]:
            params["http_headers"] = {**self.config["EXTRA_HEADERS"], **params["http_headers"]}
        tools = []
        if self.enable_functions:
            tools.extend(get_openai_schemas())
        if self.enable_mcp:
            try:
                tools.extend(get_openai_mcp_tools())
            except MCPToolsError as e:
                self.console.print(e, style="red")
        if tools:
            params["tools"] = tools
            params["tool_choice"] = "auto"
            params["parallel_tool_calls"] = False

        # Apply exclude params filtering
        params = Provider.filter_excluded_params(
            params,
            self.config,
            verbose=self.verbose,
            console=self.console,
        )

        return params

    def completion(self, messages: List[ChatMessage], stream: bool = False) -> Generator[LLMResponse, None, None]:
        """Completion method for Mistral

        Args:
            messages: List of ChatMessage
            stream: Whether to stream the response
        """
        # Convert messages to Mistral format
        mistral_messages = self._convert_messages(messages)
        if self.verbose:
            self.console.print("Messages:")
            self.console.print(mistral_messages)

        params = self.get_completion_params()
        params["messages"] = mistral_messages

        if stream:
            response = self.client.chat.stream(**params)
            yield from self._handle_stream_response(response)
        else:
            response = self.client.chat.complete(**params)
            yield from self._handle_normal_response(response)

    def _handle_normal_response(self, response: ChatCompletionResponse) -> Generator[LLMResponse, None, None]:
        """Handle normal (non-streaming) response"""
        if not response.choices or not response.choices[0].message:
            content = response.model_dump_json()
            yield LLMResponse(content=content, finish_reason="stop")
            return

        choice = response.choices[0]
        finish_reason = choice.finish_reason
        content = self.get_content_from_delta_content(choice.message.content) if choice.message.content else ""
        tool_call: Optional[ToolCall] = None

        # Handle tool calls if present
        if finish_reason == "tool_calls":
            tool_call = self._process_tool_call_chunk(choice.message.tool_calls or [])

        yield LLMResponse(content=content, finish_reason=finish_reason, tool_call=tool_call)

    def _handle_stream_response(self, response: EventStream[CompletionEvent]) -> Generator[LLMResponse, None, None]:
        """Handle stream response"""
        tool_call: Optional[ToolCall] = None

        for chunk in response:
            choice = chunk.data.choices[0]
            finish_reason = choice.finish_reason
            delta_content = choice.delta.content
            delta = choice.delta
            content = ""

            if delta_content:
                content = self.get_content_from_delta_content(delta_content)

            # Process tool call information
            if hasattr(delta, "tool_calls") and delta.tool_calls:
                tool_call = self._process_tool_call_chunk(delta.tool_calls, tool_call)

            yield LLMResponse(
                content=content,
                finish_reason=finish_reason,
                tool_call=tool_call if finish_reason == "tool_calls" else None,
            )

    def _process_tool_call_chunk(
        self, tool_calls: List[MistralToolCall], existing_tool_call: Optional[ToolCall] = None
    ) -> Optional[ToolCall]:
        """Process tool call data from a response chunk"""
        # Initialize tool call object if this is the first chunk with tool call data
        if existing_tool_call is None and tool_calls:
            tool = tool_calls[0]
            existing_tool_call = ToolCall(tool.id or gen_tool_call_id(), tool.function.name, "")

        # Accumulate arguments from multiple chunks
        if existing_tool_call:
            for tool in tool_calls:
                if not tool.function:
                    continue
                # Ensure arguments is a string
                tool_args = tool.function.arguments
                if not isinstance(tool_args, str):
                    tool_args = json.dumps(tool_args)
                existing_tool_call.arguments += tool_args

        return existing_tool_call

    def get_content_from_delta_content(self, delta_content: Union[str, List[ContentChunk]]) -> str:
        """Get content from a delta content

        If the delta content is a string, it will be returned as is.
        If the delta content is a list of ContentChunk, it will be converted to a string.
        Args:
            delta_content: Union[str, List[ContentChunk]]
        Returns:
            str
        """
        if isinstance(delta_content, str):
            return delta_content
        return self.extract_contents_list(delta_content)

    def extract_contents_list(self, delta_content: List[ContentChunk]) -> str:
        """Extract content from a list of ContentChunk

        If the content is a list of ContentChunk, it will be converted to a string.
        Args:
            delta_content: List[ContentChunk]
        Returns:
            str
        """
        content = ""
        for i in delta_content:
            _type = getattr(i, "type", None) or getattr(i, "TYPE", None)
            if _type == "text":
                i = cast(TextChunk, i)
                content += i.text
            elif _type == "image_url":
                i = cast(ImageURLChunk, i)
                content += i.image_url if isinstance(i.image_url, str) else i.image_url.url
            elif _type == "document_url":
                i = cast(DocumentURLChunk, i)
                content += f"[{i.document_name}]({i.document_url})"
            elif _type == "reference":
                i = cast(ReferenceChunk, i)
                content += "Reference IDs: " + json.dumps(i.reference_ids)
        return content

    def detect_tool_role(self) -> str:
        return "tool"



"""
Cohere API provider implementation

This module implements Cohere provider classes for different deployment options:
- CohereProvider: Standard Cohere API
- CohereBadrockProvider: AWS Bedrock integration
- CohereSagemaker: AWS Sagemaker integration
"""

from typing import Any, Dict, Generator, List, Optional

from cohere import BedrockClientV2, ClientV2, SagemakerClientV2
from cohere.types.tool_call_v2 import ToolCallV2
from cohere.types.tool_call_v2function import ToolCallV2Function

from ...config import cfg
from ...console import get_console
from ...schemas import ChatMessage, LLMResponse, ToolCall
from ...tools import get_openai_schemas
from ..provider import Provider


class CohereProvider(Provider):
    """Cohere provider implementation based on cohere library"""

    DEFAULT_BASE_URL = "https://api.cohere.com/v2"
    CLIENT_CLS = ClientV2
    DEFAULT_MODEL = "command-a-03-2025"

    def __init__(self, config: dict = cfg, verbose: bool = False, **kwargs):
        """
        Initialize the Cohere provider

        Args:
            config: Configuration dictionary
            verbose: Whether to enable verbose logging
            **kwargs: Additional parameters passed to the client
        """
        self.config = config
        self.verbose = verbose
        self.client_params = {
            "api_key": self.config["API_KEY"],
            "timeout": self.config["TIMEOUT"],
        }
        if self.config.get("BASE_URL"):
            self.client_params["base_url"] = self.config["BASE_URL"]
        self.client = self.create_client()
        self.console = get_console()

    def create_client(self):
        """Create and return Cohere client instance"""
        if self.config.get("ENVIRONMENT"):
            self.client_params["environment"] = self.config["ENVIRONMENT"]
        return self.CLIENT_CLS(**self.client_params)

    def detect_tool_role(self) -> str:
        """Return the role name for tool response messages"""
        return "tool"

    def _convert_messages(self, messages: List[ChatMessage]) -> List[Dict[str, Any]]:
        """
        Convert a list of ChatMessage objects to a list of Cohere message dicts

        {
            "role": "tool",
            "tool_call_id": tc.id,
            "content": {
                "type": "document",
                "document": {"data": string},
            },
        }

        Args:
            messages: List of ChatMessage objects

        Returns:
            List of message dicts formatted for Cohere API
        """
        converted_messages = []
        for msg in messages:
            # Create base message
            message = {}

            # Set role always
            message["role"] = msg.role

            # Add tool calls for assistant messages
            if msg.role == "assistant" and msg.tool_calls:
                # {
                #     "role": "assistant",
                #     "tool_calls": response.message.tool_calls,
                #     "tool_plan": response.message.tool_plan,
                # }
                message["tool_calls"] = [
                    ToolCallV2(
                        id=tc.id,
                        type="function",
                        function=ToolCallV2Function(name=tc.name, arguments=tc.arguments),
                    )
                    for tc in msg.tool_calls
                ]
            else:
                # Add content for non-tool-call messages
                message["content"] = msg.content or ""

            # Add tool call ID for tool messages
            if msg.role == "tool" and msg.tool_call_id:
                message["tool_call_id"] = msg.tool_call_id

                # For tool messages, convert content to the expected document format
                if msg.content:
                    message["content"] = [{"type": "document", "document": {"data": msg.content}}]

            converted_messages.append(message)

        return converted_messages

    def _prepare_tools(self) -> Optional[List[Dict[str, Any]]]:
        """
        Prepare tools for Cohere API if enabled

        Returns:
            List of tool definitions or None if disabled
        """
        if not self.config.get("ENABLE_FUNCTIONS", False):
            return None

        tools = get_openai_schemas()
        if not tools and self.verbose:
            self.console.print("No tools available", style="yellow")
        return tools

    def _handle_streaming_response(self, response_stream) -> Generator[LLMResponse, None, None]:
        """
        Process streaming response from Cohere API

        doc: https://docs.cohere.com/v2/docs/streaming

        According to Cohere docs, there are multiple event types:
        - message-start: First event with metadata
        - content-start: Start of content block
        - content-delta: Chunk of generated text
        - content-end: End of content block
        - message-end: End of message
        - tool-plan-delta: Part of tool planning
        - tool-call-start: Start of tool call
        - tool-call-delta: Part of tool call
        - tool-call-end: End of tool call
        - citation-start/end: For citations in RAG

        Args:
            response_stream: Stream from Cohere client

        Yields:
            LLMResponse objects with content or tool calls
        """
        tool_call: Optional[ToolCall] = None
        for chunk in response_stream:
            if not chunk:
                continue

            # Handle different event types
            if chunk.type == "content-delta":
                # Text generation chunks
                content = chunk.delta.message.content.text or ""
                yield LLMResponse(content=content)

            elif chunk.type == "tool-plan-delta":
                # Tool planning - when model is deciding which tool to use: cohere.types.chat_tool_plan_delta_event_delta_message.ChatToolPlanDeltaEventDeltaMessage
                content = chunk.delta.message.tool_plan or ""
                yield LLMResponse(content=content)

            elif chunk.type == "tool-call-start":
                # Start of tool call
                tool_call_msg = chunk.delta.message.tool_calls
                tool_call = ToolCall(
                    id=tool_call_msg.id, name=tool_call_msg.function.name, arguments=tool_call_msg.function.arguments
                )
                # Tool call started, waiting for tool-calls-delta events
                continue
            elif chunk.type == "tool-call-delta":
                # Tool call arguments being generated: cohere.types.chat_tool_call_delta_event_delta_message.ChatToolCallDeltaEventDeltaMessage
                if not tool_call:
                    continue
                tool_call.arguments += chunk.delta.message.tool_calls.function.arguments or ""
                # Waiting for tool-call-end event
                continue

            elif chunk.type == "tool-call-end":
                # End of a tool call, empty chunk
                yield LLMResponse(tool_call=tool_call)

    def _handle_normal_response(self, response) -> Generator[LLMResponse, None, None]:
        """
        Process non-streaming response from Cohere API

        Args:
            response: Response from Cohere client

        Yields:
            LLMResponse objects with content or tool calls
        """
        # Handle content
        if response.message.content:
            for content_item in response.message.content:
                if hasattr(content_item, "text") and content_item.text:
                    yield LLMResponse(content=content_item.text)

        # Handle tool calls
        if response.message.tool_calls:
            yield LLMResponse(content=response.message.tool_plan)
            for tool_call in response.message.tool_calls:
                yield LLMResponse(
                    tool_call=ToolCall(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        arguments=tool_call.function.arguments,
                    )
                )

    def completion(
        self, messages: List[ChatMessage], stream: bool = False, **kwargs
    ) -> Generator[LLMResponse, None, None]:
        """
        Get completion from Cohere models

        Args:
            messages: List of messages for the conversation
            stream: Whether to stream the response
            **kwargs: Additional parameters to pass to the Cohere client

        Yields:
            LLMResponse objects with content or tool calls
        """
        # Get configuration values
        model = self.config.get("MODEL", self.DEFAULT_MODEL)
        temperature = float(self.config.get("TEMPERATURE", 0.3))
        frequency_penalty = float(self.config.get("FREQUENCY_PENALTY", 0.0))
        # Prepare messages and tools
        cohere_messages = self._convert_messages(messages)
        if self.verbose:
            self.console.print("Messages:")
            self.console.print(cohere_messages)
        tools = self._prepare_tools()

        # Common request parameters
        request_params = {
            "model": model,
            "messages": cohere_messages,
            "temperature": temperature,
            "frequency_penalty": frequency_penalty,
            **kwargs,
        }

        # Add tools if available
        if tools:
            request_params["tools"] = tools

        # Apply exclude params filtering
        request_params = Provider.filter_excluded_params(
            request_params,
            self.config,
            verbose=self.verbose,
            console=self.console,
        )

        # Call Cohere API
        try:
            if stream:
                # Streaming mode
                response_stream = self.client.chat_stream(**request_params)
                yield from self._handle_streaming_response(response_stream)
            else:
                # Non-streaming mode
                response = self.client.chat(**request_params)
                yield from self._handle_normal_response(response)

        except Exception as e:
            error_msg = f"Error in Cohere API call: {e}"
            if self.verbose:
                import traceback

                self.console.print("Error in Cohere completion:")
                traceback.print_exc()
            yield LLMResponse(content=error_msg)


class CohereBadrockProvider(CohereProvider):
    """Cohere provider for AWS Bedrock integration"""

    CLIENT_CLS = BedrockClientV2
    DOC_URL = "https://docs.cohere.com/v2/docs/text-gen-quickstart"
    CLIENT_KEYS = (
        ("AWS_REGION", "aws_region"),
        ("AWS_ACCESS_KEY_ID", "aws_access_key"),
        ("AWS_SECRET_ACCESS_KEY", "aws_secret_key"),
        ("AWS_SESSION_TOKEN", "aws_session_token"),
    )

    def create_client(self):
        """Create Bedrock client with AWS credentials"""
        for k, p in self.CLIENT_KEYS:
            v = self.config.get(k, None)
            if v is None:
                raise ValueError(
                    f"You have to set key `{k}` to use {self.__class__.__name__}, see cohere doc `{self.DOC_URL}`"
                )
            self.client_params[p] = v
        return self.CLIENT_CLS(**self.client_params)


class CohereSagemakerProvider(CohereBadrockProvider):
    """Cohere provider for AWS Sagemaker integration"""

    CLIENT_CLS = SagemakerClientV2



import subprocess

from yaicli.function_schema import OpenAISchema
from pydantic import Field


class Function(OpenAISchema):
    """
    Execute a shell command and return the output (result).
    """

    shell_command: str = Field(
        ...,
        json_schema_extra={
            "example": "ls -la",
        },
        description="Shell command to execute.",
    )

    class Config:
        title = "execute_shell_command"

    @classmethod
    def execute(cls, shell_command: str) -> str:
        """
        Execute a shell command and return the output (result).

        Args:
            shell_command (str): shell command to execute.

        Returns:
            str: exit code and output string.
        """
        # Optional security check
        dangerous_commands = ["rm -rf", "mkfs", "dd"]
        if any(cmd in shell_command for cmd in dangerous_commands):
            return "Error: Dangerous command detected."

        try:
            process = subprocess.Popen(
                shell_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
            )
            output, _ = process.communicate()
            exit_code = process.returncode
            return f"Exit code: {exit_code}, Output:\n{output}"
        except Exception as e:
            return f"Error: {str(e)}"



"""Command handler for processing special commands in YAICLI."""

import shlex
import subprocess
from typing import Union

from rich.markdown import Markdown
from rich.padding import Padding

from .config import cfg
from .const import (
    CHAT_MODE,
    CMD_ADD,
    CMD_CLEAR,
    CMD_CONTEXT,
    CMD_DELETE_CHAT,
    CMD_EXIT,
    CMD_HELP,
    CMD_HISTORY,
    CMD_LIST_CHATS,
    CMD_LOAD_CHAT,
    CMD_MODE,
    CMD_SAVE_CHAT,
    EXEC_MODE,
    DefaultRoleNames,
)


class CmdHandler:
    """Handler for special commands used in the CLI."""

    def __init__(self, cli):
        """Initialize command handler with reference to CLI instance.

        Args:
            cli: CLI instance to access methods and properties
        """
        self.cli = cli
        self.commands = {
            # Register special commands with their handlers
            CMD_HELP[0] if isinstance(CMD_HELP, tuple) else CMD_HELP: self.handle_help,
            "?": self.handle_help,  # support ? as help command
            CMD_EXIT: self.handle_exit,
            CMD_CLEAR: self.handle_clear,
            CMD_HISTORY: self.handle_history,
            CMD_LIST_CHATS: self.handle_list,
            "/chats": self.handle_list,  # support /chats as alias for list command
            CMD_SAVE_CHAT: self.handle_save,
            CMD_LOAD_CHAT: self.handle_load,
            CMD_DELETE_CHAT: self.handle_delete,
            CMD_MODE: self.handle_mode,
            CMD_ADD: self.handle_add_context,
            CMD_CONTEXT[0] if isinstance(CMD_CONTEXT, tuple) else CMD_CONTEXT: self.handle_context,
            "/ctx": self.handle_context,
        }

    def handle_add_context(self, command_input: str = "") -> bool:
        """Handle /add command to add file/dir to context.

        Args:
            command_input: "/add path/to/file"

        Returns:
            True to continue REPL
        """
        try:
            parts = shlex.split(command_input)
        except ValueError as e:
            self.cli.console.print(f"Error parsing command: {e}", style="red")
            return True

        if len(parts) < 2:
            self.cli.console.print(f"Usage: {CMD_ADD} <path>", style="yellow")
            return True

        path = parts[1]
        if path.startswith("@"):
            path = path[1:]
        if self.cli.verbose:
            self.cli.console.print(f"[dim][DEBUG] Adding context: {path}[/dim]")
        self.cli.context_manager.add(path)
        return True

    def handle_context(self, command_input: str = "") -> bool:
        """Handle /context (or /ctx) command.

        Subcommands:
            list: List context items
            clear: Clear context
            remove <path>: Remove item from context
            add <path>: Add item (same as /add)

        Args:
            command_input: "/context <subcmd> [args]"
        """
        try:
            parts = shlex.split(command_input)
        except ValueError as e:
            self.cli.console.print(f"Error parsing command: {e}", style="red")
            return True

        if len(parts) < 2:
            self.cli.console.print("Usage: /context <subcommand> [args]", style="yellow")
            self.cli.console.print("Subcommands: list, clear, add <path>, remove <path>", style="dim")
            # Default to list if no subcommand
            self.cli.context_manager.list_items()
            return True

        subcmd = parts[1].lower()

        if subcmd == "list":
            self.cli.context_manager.list_items()
        elif subcmd == "clear":
            self.cli.context_manager.clear()
        elif subcmd == "add":
            if len(parts) < 3:
                self.cli.console.print("Usage: /context add <path>", style="yellow")
            else:
                path = parts[2]
                if path.startswith("@"):
                    path = path[1:]
                self.cli.context_manager.add(path)
        elif subcmd in ("remove", "rm", "delete", "del"):
            if len(parts) < 3:
                self.cli.console.print("Usage: /context remove <path>", style="yellow")
            else:
                self.cli.context_manager.remove(parts[2])
        else:
            self.cli.console.print(f"Unknown sub-command: {subcmd}", style="yellow")
            self.cli.console.print("Available: list, clear, add, remove", style="dim")

        return True

    def handle_command(self, user_input: str) -> Union[bool, str]:
        """Handle special command return: True-continue loop, False-exit loop, str-non-special command

        Args:
            user_input: Raw input from the user

        Returns:
            True if command handled and should continue loop
            False if should exit loop
            user_input if not a special command
        """
        lower_input = user_input.lower().strip()

        # Check if input starts with exclamation mark (both half-width ! and full-width ！)
        if user_input.startswith("!") or user_input.startswith("！"):
            cmd = user_input[1:].strip()
            return self.handle_shell_command(cmd)

        # Check for commands with exact command names
        if lower_input in self.commands:
            return self.commands[lower_input]()

        # Check for commands with parameters (save, load, delete, mode, add, context)
        for cmd_prefix, handler in self.commands.items():
            if lower_input.startswith(cmd_prefix) and cmd_prefix in (
                CMD_SAVE_CHAT,
                CMD_LOAD_CHAT,
                CMD_DELETE_CHAT,
                CMD_MODE,
                CMD_ADD,
                CMD_CONTEXT[0] if isinstance(CMD_CONTEXT, tuple) else CMD_CONTEXT,
                "/ctx",  # Also handle the /ctx alias
            ):
                return handler(user_input)

        # Not a special command
        return user_input

    def handle_shell_command(self, cmd: str) -> bool:
        """Execute a shell command directly with current environment variables.

        Args:
            cmd: Shell command to execute

        Returns:
            True to continue the REPL loop
        """
        if not cmd:
            return True

        self.cli.console.print(f"Executing: {cmd}", style="bold green")
        try:
            subprocess.call(cmd, shell=True)
        except Exception as e:
            self.cli.console.print(f"Failed to execute command: {e}", style="red")
        return True

    def handle_help(self) -> bool:
        """Show help message.

        Returns:
            True to continue the REPL loop
        """
        self.cli.print_help()
        return True

    def handle_exit(self) -> bool:
        """Exit the REPL loop.

        Returns:
            False to exit the REPL loop
        """
        return False

    def handle_clear(self) -> bool:
        """Clear chat history.

        Returns:
            True to continue the REPL loop
        """
        if self.cli.current_mode == CHAT_MODE:
            self.cli.chat.history.clear()
            self.cli.console.print("Chat history cleared", style="bold yellow")
        return True

    def handle_history(self) -> bool:
        """Show chat history.

        Returns:
            True to continue the REPL loop
        """
        if not self.cli.chat.history:
            self.cli.console.print("History is empty.", style="yellow")
        else:
            self.cli.console.print("Chat History:", style="bold underline")
            message_count = 1
            for msg in self.cli.chat.history:
                if msg.role == "user" and msg.content:
                    self.cli.console.print(f"[dim]{message_count}[/dim] [bold blue]User:[/bold blue] {msg.content}")
                    message_count += 1
                elif msg.role == "assistant" and (msg.content or msg.tool_calls):
                    content = msg.content or ""
                    if msg.tool_calls:
                        content += "\n" if content else ""  # Add newline if there's assistant content before tool calls
                        for t in msg.tool_calls:
                            content += f">Tool Call: {t.name}({t.arguments})\n"
                    md = Markdown(content, code_theme=cfg["CODE_THEME"])
                    padded_md = Padding(md, (0, 0, 0, 4))
                    self.cli.console.print("    Assistant:", style="bold green")
                    self.cli.console.print(padded_md)
                # Skip other roles (like "tool") and empty messages - they are not displayed
        return True

    def handle_list(self) -> bool:
        """List saved chats.

        Returns:
            True to continue the REPL loop
        """
        self.cli._list_chats()
        return True

    def handle_save(self, command_input: str = "") -> bool:
        """Save current chat.

        Args:
            command_input: Raw command input that may contain a title

        Returns:
            True to continue the REPL loop
        """
        parts = command_input.split(maxsplit=1)
        title = parts[1] if len(parts) > 1 else self.cli.chat.title
        self.cli._save_chat(title)
        return True

    def handle_load(self, command_input: str = "") -> bool:
        """Load saved chat.

        Args:
            command_input: Raw command input that should contain an index

        Returns:
            True to continue the REPL loop
        """
        parts = command_input.split(maxsplit=1)
        if len(parts) == 2 and parts[1].isdigit():
            self.cli._load_chat_by_index(index=parts[1])
        else:
            self.cli.console.print(f"Usage: {CMD_LOAD_CHAT} <index>", style="yellow")
            self.cli._list_chats()
        return True

    def handle_delete(self, command_input: str = "") -> bool:
        """Delete saved chat.

        Args:
            command_input: Raw command input that should contain an index

        Returns:
            True to continue the REPL loop
        """
        parts = command_input.split(maxsplit=1)
        if len(parts) == 2 and parts[1].isdigit():
            self.cli._delete_chat_by_index(index=parts[1])
        else:
            self.cli.console.print(f"Usage: {CMD_DELETE_CHAT} <index>", style="yellow")
            self.cli._list_chats()
        return True

    def handle_mode(self, command_input: str = "") -> bool:
        """Switch between chat and exec modes.

        Args:
            command_input: Raw command input that should contain the mode

        Returns:
            True to continue the REPL loop
        """
        parts = command_input.lower().split(maxsplit=1)
        if len(parts) == 2 and parts[1] in [CHAT_MODE, EXEC_MODE]:
            new_mode = parts[1]
            if self.cli.current_mode != new_mode:
                self.cli.current_mode = new_mode
                self.cli.set_role(DefaultRoleNames.SHELL if self.cli.current_mode == EXEC_MODE else self.cli.init_role)
            else:
                self.cli.console.print(f"Already in {self.cli.current_mode} mode.", style="yellow")
        else:
            self.cli.console.print(f"Usage: {CMD_MODE} {CHAT_MODE}|{EXEC_MODE}", style="yellow")
        return True



from typing import Any, Dict, List

from cerebras.cloud.sdk import Cerebras

from .openai_provider import OpenAIProvider


class CerebrasProvider(OpenAIProvider):
    """Cerebras LLM provider"""

    CLIENT_CLS = Cerebras
    DEFAULT_BASE_URL = "https://api.cerebras.ai"

    COMPLETION_PARAMS_KEYS = {
        "model": "MODEL",
        "temperature": "TEMPERATURE",
        "top_p": "TOP_P",
        "max_completion_tokens": "MAX_TOKENS",
        "timeout": "TIMEOUT",
        "extra_body": "EXTRA_BODY",
    }

    def get_client_params(self) -> Dict[str, Any]:
        client_params = super().get_client_params()
        client_params["warm_tcp_connection"] = False
        return client_params

    def get_tools(self) -> List[dict]:
        tools = super().get_tools()
        for i in tools:
            if "function" not in i:
                continue
            if "parameters" not in i["function"]:
                continue
            if "properties" not in i["function"]["parameters"]:
                continue
            if not isinstance(i["function"]["parameters"]["properties"], dict):
                continue
            for v in i["function"]["parameters"]["properties"].values():
                v.pop("example", None)
        return tools



import json

from yaicli.function_schema import OpenAISchema
from pydantic import Field


class Function(OpenAISchema):
    """
    Search the web using DuckDuckGo and return results as JSON.
    """

    query: str = Field(
        ...,
        json_schema_extra={
            "example": "Python programming tutorial",
        },
        description="Search query string.",
    )
    max_results: int = Field(
        default=10,
        json_schema_extra={
            "example": 10,
        },
        description="Maximum number of search results to return (default: 10).",
    )
    region: str = Field(
        default="us-en",
        json_schema_extra={
            "example": "zh-cn",
        },
        description="Region for search results (default: us-en, options: zh-cn, us-en, uk-en, etc.).",
    )
    safesearch: str = Field(
        default="moderate",
        json_schema_extra={
            "example": "moderate",
        },
        description="Safe search level: on, moderate, or off (default: moderate).",
    )

    class Config:
        title = "web_search"

    @classmethod
    def execute(cls, query: str, max_results: int = 10, region: str = "us-en", safesearch: str = "moderate") -> str:
        """
        Search the web using DuckDuckGo and return results as JSON.

        Args:
            query: Search query string.
            max_results: Maximum number of results to return.
            region: Region code for localized results (e.g., us-en, zh-cn).
            safesearch: Safe search level (on, moderate, off).

        Returns:
            str: JSON string with search results.
        """
        result = {
            "query": query,
            "engine": "duckduckgo",
            "success": False,
            "results": [],
            "result_count": 0,
            "error": None,
        }

        try:
            # Import here to avoid dependency issues if not installed
            try:
                from ddgs import DDGS
            except ImportError:
                result["error"] = "ddgs library not installed. Please install with: pip install -U ddgs"
                return json.dumps(result, ensure_ascii=False, indent=2)

            # Perform search using new ddgs API
            ddgs = DDGS()
            search_results = ddgs.text(
                query=query,
                region=region,
                safesearch=safesearch,
                max_results=max_results,
            )

            # Process results
            results = []
            for item in search_results:
                # ddgs returns dict with keys: title, href, body
                results.append(
                    {
                        "title": item.get("title", ""),
                        "url": item.get("href", ""),
                        "snippet": item.get("body", ""),
                    }
                )

            result["success"] = True
            result["results"] = results
            result["result_count"] = len(results)

            return json.dumps(result, ensure_ascii=False, indent=2)

        except Exception as e:
            result["error"] = str(e)
            return json.dumps(result, ensure_ascii=False, indent=2)



from typing import Any, Dict

from volcenginesdkarkruntime import Ark

from .openai_provider import OpenAIProvider


class DoubaoProvider(OpenAIProvider):
    """Doubao provider implementation based on openai-compatible API"""

    DEFAULT_BASE_URL = "https://ark.cn-beijing.volces.com/api/v3"
    CLIENT_CLS = Ark

    def get_client_params(self) -> Dict[str, Any]:
        # Initialize client params
        client_params = {"base_url": self.DEFAULT_BASE_URL}
        if self.config.get("API_KEY", None):
            client_params["api_key"] = self.config["API_KEY"]
        if self.config.get("BASE_URL", None):
            client_params["base_url"] = self.config["BASE_URL"]
        if self.config.get("AK", None):
            client_params["ak"] = self.config["AK"]
        if self.config.get("SK", None):
            client_params["sk"] = self.config["SK"]
        if self.config.get("REGION", None):
            client_params["region"] = self.config["REGION"]
        return client_params

    def get_completion_params(self) -> Dict[str, Any]:
        params = {
            "model": self.config["MODEL"],
            "temperature": self.config["TEMPERATURE"],
            "top_p": self.config["TOP_P"],
            "max_tokens": self.config["MAX_TOKENS"],
            "timeout": self.config["TIMEOUT"],
            "frequency_penalty": self.config.get("FREQUENCY_PENALTY", 0.0),
        }
        if self.config.get("EXTRA_BODY", None):
            params["extra_body"] = self.config["EXTRA_BODY"]
        return params
